import textwrap
import warnings
from datetime import datetime
from pathlib import Path
from typing import List, Optional, Union, Dict
from urllib.parse import urlparse

import click

from grid.openapi import V1GetDatastoreResponse, V1DatastoreOptimizationStatus, V1DatastoreSourceType
from grid.openapi.rest import ApiException
from grid.sdk import env
from grid.sdk._gql.queries import get_user_teams
from grid.sdk.affirmations import affirm, is_not_deleted, is_not_created, is_not_shallow, is_created
from grid.sdk.client import create_swagger_client
from grid.sdk.rest import GridRestClient
from grid.sdk.rest.datastores import (
    get_datastore_from_id,
    get_datastore_from_name,
    delete_datastore,
    create_datastore,
    get_datastore_list,
    mark_datastore_upload_complete,
)
from grid.sdk.user import User, user_from_logged_in_account
from grid.sdk.user_messaging import errors
from grid.sdk.utilities import SPECIAL_NAME_TO_SKIP_OBJECT_INIT
from grid.sdk.utils.datastore_uploads import (
    resume_datastore_upload,
    begin_new_datastore_upload,
    find_incomplete_datastore_upload,
    load_datastore_work_state,
)

DATASTORE_TERMINAL_STATES = [
    V1DatastoreOptimizationStatus.DELETED,
    V1DatastoreOptimizationStatus.SUCCEEDED,
    V1DatastoreOptimizationStatus.FAILED,
]

DATASTORE_FSX_THROUGHPUT_ALIASES = {
    "low": 125,
    "medium": 500,
    "high": 1000,
}


def _datastore_optimization_status_name_to_human(phase: str) -> str:
    """Converts datastore status enum name to cannoniacl format

    Example
    -------
    "DATASTORE_OPTIMIZATION_STATUS_FAILED" -> "failed"

    Notes
    -----
    See PRD for terms used: https://www.notion.so/gridai/Workload-status-observability-0b635c62db9644499d5b5fbd992385d3
    """
    conversion_map = {
        V1DatastoreOptimizationStatus.DELETED:
        "deleted",
        V1DatastoreOptimizationStatus.FAILED:
        "failed",
        V1DatastoreOptimizationStatus.SUCCEEDED:
        "available",
        V1DatastoreOptimizationStatus.OPTIMIZING:
        "optimizing",
        V1DatastoreOptimizationStatus.POD_PENDING:
        "optimizing",
        V1DatastoreOptimizationStatus.CLOUD_RESOURCES_PENDING:
        "optimizing",
        # yes, this doesn't make sense... but we need to make the datastore proto
        # spec compliant with every other resource before we can change this
        V1DatastoreOptimizationStatus.UNSPECIFIED:
        "uploading",
    }

    if phase not in conversion_map:
        warnings.warn((
            f"The operation completed, but you've found a bug! We cannot convert optimization "
            f"status name: {phase} to a human readable format. Please report this to the grid "
            f"support team on the community slack or at support@grid.ai."
        ), RuntimeWarning)
    else:
        phase = conversion_map[phase]
    return phase


def fetch_datastore(datastore_name: str, datastore_version: int, cluster: str) -> 'Datastore':
    """
    Validate the existence of provided datastore and user's access. Inject datastore id to the
    config, based on the name and version provided. If version is not provided, this function also
    injects the maximum version to the config
    """
    split = datastore_name.split(":")
    owner = None
    if len(split) == 1:
        datastore_name = split[0]
    elif len(split) == 2:
        datastore_name = split[1]
        owner = split[0]
    elif len(split) > 2:
        raise ValueError(f"Error while parsing {datastore_name}. Use the format <username>:<datastore-name>")

    # Fetching all datastores and filter them based on the arguments
    all_datastores = list_datastores(is_global=True, cluster_id=cluster)
    possible_datastores = [d for d in all_datastores if d.name == datastore_name]
    if datastore_version:
        possible_datastores = [d for d in possible_datastores if d.version == datastore_version]
    if not owner:
        # TODO - this is a hack that must be fixed after proper RBAC can fetch the datastore in a team
        user = user_from_logged_in_account()
        owner = user.username
    possible_datastores = [d for d in possible_datastores if d.user.username == owner]
    if cluster:
        possible_datastores = [d for d in possible_datastores if d.cluster_id == cluster]

    # Throwing if no datastores found
    if len(possible_datastores) == 0:
        raise ValueError(
            f'No ready-to-use datastore found with name {datastore_name} '
            f'and version {datastore_version} in the cluster {cluster}'
        )

    # choosing the latest datastore if no version is provided
    if datastore_version is None:
        selected_dstore = possible_datastores[0]
        for dstore in possible_datastores:
            if dstore.version > selected_dstore.version:
                selected_dstore = dstore
        warnings.warn(
            f'No ``--datastore_version`` passed. Using datastore: {datastore_name} version: {selected_dstore.version}'
        )
    else:
        selected_dstore = possible_datastores[0]

    return selected_dstore


class Datastore:
    _name: str
    _id: str
    _version: int
    _source: Optional[Union[str, Path]]
    _s3_no_copy: bool
    _status: Optional[str]
    _created_at: datetime
    _user: User
    _cluster_id: str
    _size: str
    _fsx_enabled: bool
    _fsx_throughput_mbs_tib: int
    _fsx_capacity_gib: int
    _fsx_preloading: bool

    _is_deleted: bool
    _is_created: bool
    _is_shallow: bool

    def __init__(
        self,
        name: Optional[str] = None,
        source: Optional[Union[str, Path]] = None,
        s3_no_copy: bool = False,
        user: Optional[User] = None,
        version: int = 0,
        cluster_id: Optional[str] = None,
        fsx_enabled: bool = False,
        fsx_throughput_alias="low",
        fsx_capacity_gib=1200,
        fsx_preloading=False,
    ):
        """Initialize a new DataStore Object.

        If a DataStore with the given name, version, team and cluster already exists,
        then the object returned will be able to interact with the existing DataStore.

        Alternatively, if the DataStore is going to be created for the first time, then
        the ``source` parameters can be used to specify the location of the DataStore on
        disk (or at a remote location).

        After initializing the datastore object, the data itself can be uploaded by calling
        the ``upload()`` method.

        TODO
        ----
        - user and team shouldn't be arguments

        Parameters
        ----------
        name
            The name of the DataStore.
        version
            The version of the DataStore.
        source
            The location of the DataStore on disk or at a remote location.
        s3_no_copy
            If using an s3:// bucket as source use a remote reference type
        user
            The user that owns the DataStore.
        cluster_id
            The name of the cluster that the DataStore should be uploaded to.
        fsx_enabled
            Whether to provision an FSx-backed datastore
        fsx_throughput_alias
            Which throughput tier to provision for the FSx file system. low=125mb/s/tib, medium=500mb/s/tib, high=1000mb/s/tib
        fsx_capacity_gib
            How much capacity to provision for the FSx file system, must be 1200, 2400 or a multiple of 2400
        fsx_preloading
            Whether to preload the data to the FSx file system on datastore creation
        """
        # --------------------------------------------------------------------------- #
        #    This should be the first block that goes into the constructor of the     #
        #    resource object. This block sets the correct values for the private      #
        #    attributes which is then later picked up by the decorator(s) to take     #
        #    right actions. It also initialize the _client object and cluster_id      #
        #    which is required by the downstream methods regardless of the object     #
        #    is completely initialized or not. Most importantly, this blocks checks   #
        #    for the name argument to decide if it's a call from other internal       #
        #    methods to create a shallow object. This is done by checking the         #
        #    special name variable. Other methods that already has the backend        #
        #    response fetched, can use this to create the object without the backend  #
        #    call and then fill-in the response they already have.                    #
        #                                                                             #
        self._client = GridRestClient(api_client=create_swagger_client())
        cluster_id = cluster_id or env.CONTEXT
        self._is_shallow = False
        self._cluster_id = cluster_id
        if name == SPECIAL_NAME_TO_SKIP_OBJECT_INIT:
            self._is_shallow = True
            self._is_created = False
            self._is_deleted = False
            return
        #                                                                             #
        # --------------------------------------------------------------------------- #

        if name is None:
            if source:
                name = parse_name_from_source(source)
            else:
                raise ValueError("Name is required if source is not provided.")
        else:
            try:

                datastore = get_datastore_from_name(
                    client=self._client, cluster_id=cluster_id, datastore_name=name, version=version
                )
                self.__dict__ = self._setup_from_response(datastore).__dict__
                return
            except KeyError:
                self._is_deleted = False  # the datastore has not been deleted
                self._is_created = False  # it doesn't exists in the grid backend.
                pass
        if version:
            raise ValueError(
                f"Existing datastore with name {name} and version {version} is not found. "
                f"If you are creating a new datastore, avoid passing a version argument "
                f"as this is auto-generated."
            )

        self._fsx_throughput_mbs_tib = 0
        self._fsx_capacity_gib = 0
        self._fsx_preloading = False

        if fsx_enabled is True:
            if fsx_throughput_alias not in DATASTORE_FSX_THROUGHPUT_ALIASES:
                raise ValueError(
                    f"Invalid throughput {fsx_throughput_alias} provided for high performance datastore."
                    f"Please use one of {DATASTORE_FSX_THROUGHPUT_ALIASES.keys()}."
                )
            else:
                self._fsx_throughput_mbs_tib = DATASTORE_FSX_THROUGHPUT_ALIASES[fsx_throughput_alias]
                self._fsx_capacity_gib = fsx_capacity_gib
                self._fsx_preloading = fsx_preloading

        self._name = name
        self._version = 0
        self._source = source
        self._user = user
        self._cluster_id = cluster_id
        self._s3_no_copy = s3_no_copy
        self._fsx_enabled = fsx_enabled
        self._id = None
        self._status = None
        self._created_at = None
        self._size = None

        try:
            resp = get_datastore_from_name(
                client=self._client, cluster_id=self._cluster_id, datastore_name=self._name, version=self._version
            )
            self.__dict__ = self._setup_from_response(resp).__dict__
        except KeyError:
            self._is_deleted = False
            self._is_created = False
            self._is_shallow = False

    @classmethod
    def _setup_from_response(cls, datastore: V1GetDatastoreResponse) -> 'Datastore':
        instance = cls(name=SPECIAL_NAME_TO_SKIP_OBJECT_INIT)
        instance._is_deleted = datastore.status.phase == V1DatastoreOptimizationStatus.DELETED
        instance._is_created = True
        instance._is_shallow = False

        instance._id = datastore.id
        instance._name = datastore.name
        instance._cluster_id = datastore.spec.cluster_id
        instance._version = datastore.spec.version
        instance._source = datastore.spec.source
        instance._s3_no_copy = instance._source == V1DatastoreSourceType.OBJECT_STORE_REFERENCE_ONLY
        instance._fsx_enabled = instance._source == V1DatastoreSourceType.FSX
        if instance._fsx_enabled:
            instance._fsx_throughput_mbs_tib = datastore.spec.fsx_spec.storage_throughput_mb_s_tib
            instance._fsx_capacity_gib = datastore.spec.fsx_spec.storage_capacity_gib
            instance._fsx_preloading = datastore.spec.fsx_spec.preload_data_on_create
        else:
            instance._fsx_throughput_mbs_tib = 0
            instance._fsx_capacity_gib = 0
            instance._fsx_preloading = False
        instance._status = datastore.status.phase
        instance._created_at = datastore.created_at
        instance._size = f"{datastore.spec.size_mib} MiB"
        instance._user = User(user_id=datastore.spec.user_id, username="", first_name="", last_name="")
        return instance

    @classmethod
    def _from_id(cls, datastore_id: str, cluster_id: Optional[str] = env.CONTEXT) -> "Datastore":
        instance = cls(name=SPECIAL_NAME_TO_SKIP_OBJECT_INIT, cluster_id=cluster_id)
        instance._id = datastore_id
        instance._is_shallow = True
        return instance

    @property
    def id(self) -> str:
        return self._id

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def s3_no_copy(self):
        """When using an s3:// bucket source, this flag determines if we the data should
        be copied into the grid managed bucket, or if we should only keep a reference to it.
        """
        return self._s3_no_copy

    # ------------------ Attributes Only Valid Before Upload ---------------

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def source(self) -> Union[str, Path]:
        """The directory path at which the datastore is initialized from.

        !!! Note

            This property is only available to the instance of this class which uploads
            the datastore. Previously existing datastores will not possess any value
            for this property.
        """
        return self._source

    # ------------------ Attributes Fully Active After Upload ---------------

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def name(self) -> str:
        """The name of the datastore.
        """
        return self._name

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def version(self) -> int:
        """The version of the datastore.
        """
        return self._version

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def user(self) -> User:
        """Information about the owner of the datastore (name, username, etc).
        """
        return self._user

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def created_at(self) -> datetime:
        """Date-Time timestamp when this datastore was created (first uploaded).
        """
        return self._created_at

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def size(self) -> str:
        """Size (in Bytes) of the datastore.
        """
        return self._size

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def status(self) -> str:
        """The status of the datastore.
        """
        if (self._status and (self._status not in DATASTORE_TERMINAL_STATES)) or ((self._status is None) and
                                                                                  (self._is_created is True)):
            self._update_status()

        return _datastore_optimization_status_name_to_human(self._status)

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def cluster_id(self) -> str:
        """ID of the cluster which this datastore is uploaded to.

        !!! info

            This feature is only available to bring-your-own-cloud-credentials
            customers. Please see https://www.grid.ai/pricing/ for more info.
        """
        return self._cluster_id

    def _unshallow(self):
        """ If the object is a shallow (i.e. only has an id and `_is_shallow` attribute is True)
        object, this method can be triggered to get the full object from the BE. It is designed
        to be called only from the `is_not_shallow` decorator and should not be called directly.
        """
        if not self._is_shallow:
            raise RuntimeError('Datastore is already unshallow')
        if not hasattr(self, '_id') or self._id is None:
            raise RuntimeError("Cannot unshallow resource without a valid Datastore id")
        self._is_shallow = False
        try:
            datastore = get_datastore_from_id(self._client, datastore_id=self._id, cluster_id=self.cluster_id)
        except ApiException as e:  # TODO change to GridException
            if hasattr(e, 'reason') and e.reason == "Not Found":
                self._is_deleted = True
                self._status = V1DatastoreOptimizationStatus.DELETED
        else:
            self.__dict__ = self._setup_from_response(datastore).__dict__

    # -------------- Dunder Methods ----------------------

    @affirm(is_not_shallow, is_not_deleted)
    def __repr__(self):
        if self._is_created:
            res = textwrap.dedent(
                f"""\
                {self.__class__.__name__}(
                    {"name": <10} = \"{self.name}\",
                    {"version": <10} = {self.version},
                    {"size": <10} = \"{self.size}\",
                    {"created_at": <10} = {self.created_at},
                    {"owner": <10} = {self.user},
                    {"cluster_id": <10} = {self.cluster_id},
                )"""
            )
        else:
            res = textwrap.dedent(
                f"""\
                {self.__class__.__name__}(
                    {"name": <12} = \"{self.name}\",
                    {"version": <12} = {self.version},
                    {"source": <12} = \"{self.source}\",
                    {"s3_no_copy": <12} = {self.s3_no_copy},
                    {"owner": <12} = {self.user},
                    {"cluster_id": <12} = {self.cluster_id},
                )"""
            )
        return res

    @affirm(is_not_shallow, is_not_deleted)
    def __str__(self):
        return repr(self)

    @affirm(is_not_shallow, is_not_deleted)
    def __eq__(self, other: 'Datastore'):
        # TODO - handling team's datastore equality here is probably not the best. We should
        #  delegate that to the backend when Project lands
        # need to handle case where attributes of a DataStore are not `User` or `Team`
        # classes. This is the case before the datastore is uploaded.
        self_owner = self._user.user_id if hasattr(self._user, 'user_id') else self._user
        other_owner = other._user.user_id if hasattr(other._user, 'user_id') else other.user

        return (
            self.__class__.__qualname__ == other.__class__.__qualname__ and self._name == other._name
            and self._version == other._version and self_owner == other_owner
        )

    @affirm(is_not_shallow, is_not_deleted)
    def __hash__(self):
        return hash((
            self._name, self._id, self._version, self._size, self._created_at, self._status, self._user, self._source,
            self._cluster_id, self._is_deleted, self._is_created
        ))

    # ---------------------  Internal Methods ----------------------

    @affirm(is_not_shallow, is_not_deleted)
    def _update_status(self):
        """Refreshes the``snapshot_status`` attribute by querying the Grid API.
        """
        updated = get_datastore_from_id(c=self._client, cluster_id=self._cluster_id, datastore_id=self._id)
        self._status = str(updated.status.phase)

    # ---------------------  Public Facing Methods ---------------------

    @affirm(is_not_shallow, is_not_deleted, is_created)
    def delete(self):
        """Deletes the datastore from the grid system.
        """
        delete_datastore(c=self._client, cluster_id=self._cluster_id, datastore_id=self._id)
        self._is_deleted = True

    @affirm(is_not_shallow, is_not_created)
    def upload(self):
        """Uploads the contents of the directories referenced by this datastore instance to Grid.

        Depending on your internet connection this may be a potentially long running process.
        If uploading is inturupsed, the upload session can be resumed by initializing this
        ``Datastore`` object again with the same parameters repeating the call to ``upload()``.
        """
        self._client = GridRestClient(create_swagger_client())
        dstore = create_datastore(
            c=self._client,
            cluster_id=self._cluster_id,
            name=self._name,
            source=self._source,
            no_copy_data_source=self._s3_no_copy,
            fsx_enabled=self._fsx_enabled,
            fsx_throughput_mbs_tib=self._fsx_throughput_mbs_tib,
            fsx_capacity_gib=self._fsx_capacity_gib,
            fsx_preloading=self._fsx_preloading,
        )

        self._id = dstore.id
        self._version = dstore.spec.version
        self._created_at = dstore.created_at

        if dstore.spec.source_type != V1DatastoreSourceType.EXPANDED_FILES:
            mark_datastore_upload_complete(c=self._client, cluster_id=self._cluster_id, datastore_id=self._id)
            dstore_resp = get_datastore_from_id(c=self._client, datastore_id=self._id, cluster_id=self._cluster_id)
            self.__dict__ = self._setup_from_response(dstore_resp).__dict__
            return

        incomplete_id = find_incomplete_datastore_upload(grid_dir=Path(env.GRID_DIR))
        if (incomplete_id is not None) and (incomplete_id == self._id):
            initial_work = load_datastore_work_state(grid_dir=Path(env.GRID_DIR), datastore_id=incomplete_id)
            resume_datastore_upload(client=self._client, grid_dir=Path(env.GRID_DIR), work=initial_work)
            dstore_resp = get_datastore_from_id(c=self._client, datastore_id=self._id, cluster_id=self._cluster_id)
            self.__dict__ = self._setup_from_response(dstore_resp).__dict__
            return

        begin_new_datastore_upload(
            client=self._client,
            grid_dir=Path(env.GRID_DIR),
            source_path=Path(self._source),
            cluster_id=self._cluster_id,
            datastore_id=self._id,
            datastore_name=self._name,
            creation_timestamp=self._created_at,
            datastore_version=str(self._version),
        )
        dstore_resp = get_datastore_from_id(c=self._client, datastore_id=self._id, cluster_id=self._cluster_id)
        self.__dict__ = self._setup_from_response(dstore_resp).__dict__
        return


def list_datastores(cluster_id: Optional[str] = None, is_global: bool = False) -> List[Datastore]:
    """List datastores for user / teams

    Parameters
    ----------
    is_global:
        if True, returns a list of datastores of the everyone in the team
    cluster_id:
        if specified, returns a list of datastores for the specified cluster
    """
    user = user_from_logged_in_account()
    client = GridRestClient(create_swagger_client())

    cluster_id = cluster_id or env.CONTEXT
    datastores = []

    if not is_global:
        # single user only
        user_datastore_resps = get_datastore_list(client=client, cluster_id=cluster_id)
        for user_datastore in user_datastore_resps:
            dstore_resp = get_datastore_from_id(
                c=client, datastore_id=user_datastore.id, cluster_id=user_datastore.spec.cluster_id
            )
            dstore = Datastore._setup_from_response(datastore=dstore_resp)
            dstore.user.username = user.username
            datastores.append(dstore)
    else:
        # If ``include_teams`` is set, add datastores registered to the team.
        team_user_id_name_map: Dict[str, Dict[str, str]] = {}
        team_user_id_name_map[user.user_id] = {"username": user.username, "team_name": ""}
        for team_data in get_user_teams():
            for member_data in team_data['members']:
                team_user_id_name_map[member_data['id']] = {
                    "username": member_data['username'],
                    "team_name": team_data['name'],
                }

        team_dstore_list = get_datastore_list(
            client=client, cluster_id=cluster_id, user_ids=list(team_user_id_name_map.keys())
        )
        for team_dstore in team_dstore_list:
            resp = get_datastore_from_id(c=client, cluster_id=team_dstore.spec.cluster_id, datastore_id=team_dstore.id)
            dstore = Datastore._setup_from_response(datastore=resp)
            dstore.user.username = team_user_id_name_map[dstore.user.user_id]['username']
            dstore.user.team_name = team_user_id_name_map[dstore.user.user_id]['team_name']
            datastores.append(dstore)

    return datastores


def parse_name_from_source(source) -> str:
    """Parses datastore name from source if name isn't provided"""
    try:
        parse_result = urlparse(source)
    except ValueError:
        raise click.ClickException(errors.datastore_invalid_source(source))

    if parse_result.path != '/' and parse_result.path != '':
        path = Path(parse_result.path)
        base = path.name.split(".")[0]
        ret = base.lower().strip()
    else:
        ret = parse_result.netloc

    return ret



from typing import Optional

import click

from grid.cli import rich_click, observables
from grid.sdk.rest.exceptions import GridException


@click.option(
    '--global',
    'is_global',
    type=bool,
    is_flag=True,
    help='Fetch history from everyone in the team when flag is passed'
)
@rich_click.command()
def history(is_global: Optional[bool] = False) -> None:
    """View list of historic Runs."""
    try:
        observables.Run().get_history(is_global=is_global)
    except GridException as e:
        raise click.ClickException(str(e))



import concurrent.futures
import datetime
import math
import os
import signal
import time
import warnings
from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass, field, asdict
from hashlib import blake2b
from pathlib import Path
from threading import Event
from typing import List, Optional, TYPE_CHECKING, Dict, Union

import requests
import ujson
from dataclasses_json import dataclass_json
from requests.adapters import HTTPAdapter
from rich.live import Live
from rich.panel import Panel
from rich.progress import Progress, TextColumn, BarColumn, DownloadColumn, TaskID, TransferSpeedColumn
from rich.status import Status
from urllib3 import Retry

from grid.sdk.user_messaging import errors

try:
    from rich.console import RenderGroup as Group
except ImportError:
    from rich.console import Group

from grid.metadata import __version__
from grid.sdk import env
from grid.sdk.rest.datastores import (
    datastore_upload_object_from_data_file,
    create_presigned_urls,
    complete_presigned_url_upload,
    mark_datastore_upload_complete,
)

if TYPE_CHECKING:
    from grid.sdk.rest import GridRestClient

_SEC = 1
_MIN = 60 * _SEC
_HOUR = 60 * _MIN
_DAY = 24 * _HOUR

_MAX_FILE_SIZE = 10_000 * env.MAX_BYTES_PER_FILE_PART_UPLOAD  # s3 limit is 10,000 parts

warnings.simplefilter("ignore")

# -------------- interrupt handling ---------------

done_event = Event()
received_signal = None


def handle_termination_signal(signum, frame):
    global received_signal
    done_event.set()
    received_signal = signum


def wrap_signals(fn):
    def wrapper(*args, **kwargs):

        global received_signal
        prev_sigint_handler = signal.signal(signal.SIGINT, handle_termination_signal)
        prev_sigterm_handler = signal.signal(signal.SIGTERM, handle_termination_signal)

        try:
            fn(*args, **kwargs)
        except RuntimeError as e:
            print(e)

        signal.signal(signal.SIGTERM, prev_sigterm_handler)
        signal.signal(signal.SIGINT, prev_sigint_handler)
        if received_signal is not None:
            signal.raise_signal(received_signal)

    return wrapper


# ----------------- data structures ----------------


@dataclass
class PBar:
    p_upload_bytes: Progress = Progress(
        TextColumn("[progress.description]{task.description}", justify="right"),
        BarColumn(bar_width=None),
        "[progress.percentage]{task.percentage:>3.1f}%",
        "•",
        DownloadColumn(),
        "•",
        TransferSpeedColumn(),
    )

    p_upload_files: Progress = Progress(
        TextColumn("[progress.description]{task.description}", justify="right"),
        BarColumn(bar_width=None),
        "{task.completed} / {task.total} Parts Completed",
    )

    p_finalize: Progress = Progress(
        TextColumn("[progress.description]{task.description}"),
        BarColumn(bar_width=None),
        TextColumn("{task.completed} / {task.total} File Parts"),
    )

    action_status: Status = Status("[bold yellow] Initializing Upload")

    upload_bytes_task_id: TaskID = p_upload_bytes.add_task("Upload Progress (Data Bytes)")
    upload_files_task_id: TaskID = p_upload_files.add_task("Upload Progress (File Parts)")
    finalize_task_id: TaskID = p_finalize.add_task("Finalizing Upload", visible=False)

    # group of progress bars;
    # some are always visible, others will disappear when progress is complete
    progress_group: Group = Group(
        Panel(Group(p_upload_files, p_upload_bytes, p_finalize)),
        action_status,
    )

    upload_bytes_beginning_advance: int = None
    upload_bytes_total: int = None
    upload_files_beginning_advance: int = None
    upload_files_total: int = None
    finalize_beginning_advance: int = None
    finalize_total: int = None

    status_text_initializing: str = "[bold yellow] Indexing Datastore Files"
    status_text_uploading: str = "[bold blue] Uploading Data"
    status_text_saving: str = "[bold yellow] Checkpointing Current Upload Progress"
    status_text_finalizing: str = "[bold blue] Finalizing Upload"
    status_text_done: str = "[bold green] Done!"

    def __post_init__(self):
        self.p_upload_bytes.refresh()
        self.p_upload_files.refresh()
        self.p_finalize.refresh()

    def setup_progress_meter(self, work: 'Work'):
        current_part_progress, total_part_count = 0, 0
        current_byte_progress, total_byte_count = 0, 0
        current_finalize_progress, total_finalize_files = 0, 0

        for file in work.files.values():
            total_byte_count += file.size_bytes
            total_part_count += int(file.part_count)
            total_finalize_files += 1

            if file.is_uploaded:
                current_byte_progress += file.size_bytes
                current_part_progress += int(file.part_count)
            if file.is_marked_complete:
                current_finalize_progress += 1

        self.upload_bytes_beginning_advance = current_byte_progress
        self.upload_bytes_total = total_byte_count
        self.upload_files_beginning_advance = current_part_progress
        self.upload_files_total = total_part_count
        self.finalize_beginning_advance = current_finalize_progress
        self.finalize_total = total_finalize_files

        self.p_upload_bytes.update(
            self.upload_bytes_task_id, advance=self.upload_bytes_beginning_advance, total=int(self.upload_bytes_total)
        )
        self.p_upload_files.update(
            self.upload_files_task_id, advance=self.upload_files_beginning_advance, total=self.upload_files_total
        )
        self.p_finalize.update(
            self.finalize_task_id, advance=self.finalize_beginning_advance, total=self.finalize_total, visible=False
        )


@dataclass_json
@dataclass
class UploadTask:
    data_file_local_id: str
    part_number: str
    read_offset_bytes: int  # beginning position to seek to in the file
    read_range_bytes: int  # how many bytes to read after the starting position
    url: str
    etag: Optional[str] = None


@dataclass_json
@dataclass
class DataFile:
    absolute_path: str  # path to file to upload on client
    relative_file_name: str  # key of the object in the datastore (relative path from datastore source at creation)
    size_bytes: int  # total size of the file in bytes
    local_id: str  # uuid generated when the file contents are first read.

    part_count: Optional[int] = None
    upload_id: Optional[str] = None
    expiry_time: Optional[int] = None  # unix timestamp when presigned url will expire.
    tasks: List[UploadTask] = field(default_factory=list)

    is_uploaded: bool = False
    is_marked_complete: bool = False


@dataclass_json
@dataclass
class Work:
    datastore_id: str
    datastore_name: str
    datastore_version: str
    cluster_id: str
    source: str
    creation_timestamp: int  # unix timestamp of the initial creation time
    grid_cli_version: str = __version__
    files: Dict[str, DataFile] = field(default_factory=dict)  # files contained in the datastore

    def check_for_modified_files(self) -> List[DataFile]:
        """Find all files which exist in self but which have changed or been removed from other.

        This method calculates the difference between two sets of work. For the sakke of simplicity
        we do not detect if files within the original datastore source dir have been added which were
        not included in the initial work state; nor do we report the difference type of each file
        that is different in the original work from the current state (as in, we do not report if
        a difference is due to a file being changed or if it is due to the file being removed / moved
        from the recorded filesystem path).
        """
        current_filesystem_work_state = initialize_upload_work(
            name=self.datastore_name,
            version=self.datastore_version,
            datastore_id=self.datastore_id,
            cluster_id=self.cluster_id,
            creation_timestamp=self.creation_timestamp,
            source_path=str(self.source),
        )
        differences = []
        different_in_self = set(self.files.keys()).difference(current_filesystem_work_state.files.keys())
        for difference in different_in_self:
            differences.append(self.files[difference])
        return differences


# ------------------ serialization / de-serialization --------------------


def load_datastore_work_state(grid_dir: Path, datastore_id: str) -> Work:
    state_file = grid_dir.joinpath("datastores", f"{datastore_id}.json")
    if not state_file.exists():
        raise FileNotFoundError(f"work state file does not exist at: {state_file}")
    work = Work.from_dict(ujson.loads(state_file.read_text()))
    if work.grid_cli_version != __version__:
        raise RuntimeError(
            "An incomplete datastore upload was dected which was created from a different version "
            "of the grid CLI. The incomplete work has been invalidated and the datastore must be "
            "re-uploaded in full as part of a new `grid datastore create` command."
        )
    return work


def _save_work_state(grid_dir: Path, work: Work) -> None:
    datastores_dir = grid_dir.joinpath("datastores")
    datastores_dir.mkdir(parents=True, exist_ok=True)
    state_file = datastores_dir.joinpath(f"{work.datastore_id}.json")
    state_file.touch()
    state_file.write_text(ujson.dumps(asdict(work)))


def remove_datastore_work_state(grid_dir: Path, datastore_id: str) -> None:
    state_file = grid_dir / "datastores" / f"{datastore_id}.json"
    if not state_file.exists():
        raise FileNotFoundError(f"could not find / remove work state file: {state_file}")
    os.remove(str(state_file.absolute()))


def find_incomplete_datastore_upload(grid_dir: Path) -> Union[str, None]:
    """Finds an incomplete upload in the grid dir and returns the datastore id if it exsists, else None"""
    datastores_dir = grid_dir.joinpath("datastores")
    datastores_dir.mkdir(parents=True, exist_ok=True)
    for work_state_file in list(datastores_dir.iterdir()):
        if not work_state_file.is_file() or not work_state_file.name.endswith(".json"):
            continue
        datastore_id = work_state_file.name.rstrip(".json")
        return datastore_id
    return None


# ----------------------- initialization ------------------------------

_hasher = blake2b(digest_size=15)


def _calculate_file_local_id(f: Path) -> str:
    """Calculate a deterministic / unique identifier for a file in the work index.

    This calculates a blake2b digest based off the file path, mod-time, & file size.
    This is done instead of simply generating a UUID for the key in order to allow
    incomplete file upload `Work` dicts to be compared to a resumed datastore
    upload work state simply by checking if all the keys in the dictionary are
    the same (an op resolving to a simple set difference on dict key pairs rather
    than some sort of  iteration -> sort -> attr lookup -> set comparison method)
    """
    h = _hasher.copy()
    h.update(str(f.resolve()).encode('utf-8'))
    h.update(str(f.stat().st_mtime_ns).encode('utf-8'))
    h.update(str(f.stat().st_size).encode('utf-8'))
    return h.hexdigest()


def initialize_upload_work(
    name: str,
    datastore_id: str,
    cluster_id: str,
    creation_timestamp: int,
    source_path: str,
    version: str,
) -> Work:
    rel_path = Path(source_path)
    abs_path = rel_path.absolute()
    if not abs_path.exists():
        raise OSError(f"the datastore upload source path: {source_path} does not exist")

    all_work = Work(
        datastore_name=name,
        datastore_id=datastore_id,
        cluster_id=cluster_id,
        creation_timestamp=creation_timestamp,
        source=str(abs_path),
        datastore_version=version,
    )

    # handle a single file upload case.
    if abs_path.is_file():
        local_id = _calculate_file_local_id(abs_path)
        file_size = abs_path.stat().st_size
        if file_size >= _MAX_FILE_SIZE:
            raise RuntimeError(
                f"file {abs_path.absolute()} exceeds the maximum file size for "
                f"uploads to grid datastores. please contact support @grid.ai"
                f"for assistance"
            )
        all_work.files[local_id] = DataFile(
            absolute_path=str(abs_path.resolve()),
            relative_file_name=abs_path.name,
            size_bytes=file_size,
            local_id=local_id,
            part_count=max(1, math.ceil(file_size / env.MAX_BYTES_PER_FILE_PART_UPLOAD))
        )
        return all_work

    # if the source_path is a directory...
    for f in abs_path.glob("**/*"):
        if f.is_dir():
            continue

        if f.is_symlink():
            warnings.warn(f"Cannot upload symlinked files. Skipping: {str(f)}", category=UserWarning)
            continue

        local_id = _calculate_file_local_id(f)
        file_size = f.stat().st_size
        if file_size >= _MAX_FILE_SIZE:
            raise RuntimeError(
                f"file {f.absolute()} exceeds the maximum file size for "
                f"uploads to grid datastores. please contact support @grid.ai"
                f"for assistance."
            )

        all_work.files[local_id] = DataFile(
            absolute_path=str(f.resolve()),
            relative_file_name=str(f.relative_to(abs_path)),
            size_bytes=file_size,
            local_id=local_id,
            part_count=max(1, math.ceil(file_size / env.MAX_BYTES_PER_FILE_PART_UPLOAD))
        )

    return all_work


# ----------------------- perform upload -----------------------------


def _get_next_upload_batch(work: Work) -> List[DataFile]:
    num_bytes, num_files = 0, 0
    next_batch = []
    for file in work.files.values():
        if file.is_uploaded:
            continue
        next_batch.append(file)
        num_bytes += file.size_bytes
        num_files += 1
        # ordering is important here in case there is a very large
        # files which exceeds the MAX_BYTES_PER_BATCH_UPLOAD limit.
        if num_bytes > env.MAX_BYTES_PER_BATCH_UPLOAD or num_files >= env.MAX_FILES_PER_UPLOAD_BATCH:
            break

    return next_batch


def _do_upload(session: requests.Session, file_path: Path, task: UploadTask, progress_bar: PBar) -> UploadTask:
    """Do upload, fill ETag value in UploadTask
    """
    if done_event.is_set():
        return task

    try:
        with file_path.open("rb") as file:
            file.seek(task.read_offset_bytes)
            data = file.read(task.read_range_bytes)
            bytes_read = len(data)

        response = session.put(task.url, data=data)
        if 'ETag' not in response.headers:
            raise ValueError(f"Unexpected response from S3, response: {response.content}")
        task.etag = str(response.headers['ETag']).strip('"')
    except FileNotFoundError:
        # don't error and allow the progress bar to continue updating as expected.
        bytes_read = task.read_range_bytes

    progress_bar.p_upload_bytes.update(progress_bar.upload_bytes_task_id, advance=bytes_read)
    progress_bar.p_upload_files.update(progress_bar.upload_files_task_id, advance=1)
    return task


def _upload_next_batch(c: 'GridRestClient', work: Work, progress_bar: PBar) -> Dict[str, DataFile]:
    upload_objects = []
    batch = _get_next_upload_batch(work)
    if len(batch) == 0:
        return {}

    for file in batch:
        up_obj = datastore_upload_object_from_data_file(file)
        upload_objects.append(up_obj)

    responses = create_presigned_urls(
        c=c,
        cluster_id=work.cluster_id,
        datastore_id=work.datastore_id,
        upload_objects=upload_objects,
    )

    batch_files = {}
    for resp, file in zip(responses, batch):
        file.upload_id = resp.upload_id
        file.expiry_time = resp.expiry_time
        file.part_count = resp.part_count
        for idx, url in enumerate(resp.urls):
            offset = idx * env.MAX_BYTES_PER_FILE_PART_UPLOAD
            file.tasks.append(
                UploadTask(
                    data_file_local_id=file.local_id,
                    part_number=url.part_number,
                    url=url.url,
                    read_offset_bytes=offset,
                    read_range_bytes=env.MAX_BYTES_PER_FILE_PART_UPLOAD,
                )
            )
        batch_files[file.local_id] = file

    # We can use a with statement to ensure threads are cleaned up promptly
    #
    # The algorithm dictating backoff factor is as follows:
    #
    # {backoff factor} * (2 ** ({number of total retries} - 1))
    #
    # For example, if the backoff factor is set to:
    #
    # 1 second the successive sleeps will be 0.5, 1, 2, 4, 8, 16, 32, 64, 128, 256.
    # 2 seconds - 1, 2, 4, 8, 16, 32, 64, 128, 256, 512
    # 10 seconds - 5, 10, 20, 40, 80, 160, 320, 640, 1280, 2560
    #
    # To find the total time spent before erroring, sum up backoff factors for
    # each sleep iterations up to the total number of retries (max_retries)
    retries = Retry(
        total=8,
        # status_forcelist retry codes
        # ----------------------------
        # 104: connection reset by peer can occur when 307 redirect requests aren't received quickly enough.
        # 429: too many requests (just need to slow down... exponential backoff handles this)
        # 500, 502, 503, 504: assorted errors AWS uses to say "yeah... we screwed up"
        status_forcelist=[104, 429, 500, 502, 503, 504],
        method_whitelist=["HEAD", "GET", "PUT", "POST"],
        backoff_factor=2,
    )
    max_work_threads = env.UPLOAD_MAX_WORKER_THREADS
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_work_threads) as executor, requests.Session() as s:
        # For explanation of HTTPAdapter and request session see: https://stackoverflow.com/a/34893364
        s.mount(
            'https://',
            HTTPAdapter(
                # should only ever "need" 1 pool_connection(s); but using 3 to be safe in even of redirects
                pool_connections=3,
                pool_maxsize=max_work_threads * 2,
                max_retries=retries,
            )
        )
        futures = []
        uploaded_file_tasks: Dict[str, List[UploadTask]] = defaultdict(list)
        for file in batch_files.values():
            file_path = Path(file.absolute_path)
            for task in file.tasks:
                futures.append(executor.submit(_do_upload, s, file_path, task, progress_bar))

        for future in concurrent.futures.as_completed(futures):
            if done_event.is_set():
                raise RuntimeError(errors.datastore_upload_process_cancelled())
            res = future.result()
            uploaded_file_tasks[res.data_file_local_id].append(res)

    for k in list(uploaded_file_tasks.keys()):
        uploaded_file_tasks[k] = sorted(uploaded_file_tasks[k], key=lambda x: int(x.part_number))

    completed_files: Dict[str, DataFile] = {}
    for data_file_local_id, completed_tasks in uploaded_file_tasks.items():
        completed_file = batch_files[data_file_local_id]
        completed_file.tasks = completed_tasks
        completed_file.is_uploaded = True
        completed_files[data_file_local_id] = completed_file

    return completed_files


def _start_upload(c: 'GridRestClient', grid_dir: Path, work: Work, progress_bar: PBar) -> Work:
    """Entrypoint to begin processing file uploads.
    """
    while True:
        if done_event.is_set():
            # stop progression since we're in trouble.
            raise RuntimeError(errors.datastore_upload_process_cancelled())

        progress_bar.action_status.update(progress_bar.status_text_uploading)

        completed_files = _upload_next_batch(c=c, work=work, progress_bar=progress_bar)
        if done_event.is_set():
            # don't save if we're not sure that everything completed uploading.
            raise RuntimeError(errors.datastore_upload_process_cancelled())

        if len(completed_files) == 0:
            break
        for file_id, file in completed_files.items():
            work.files[file_id] = file

        progress_bar.action_status.update(progress_bar.status_text_saving)
        _save_work_state(grid_dir, work)

    return work


# ---------------------------- perform finalize ---------------------------


def _get_next_finalize_batch(work: Work) -> List[DataFile]:
    next_batch = []
    num_tasks = 0
    for file in work.files.values():
        if file.is_marked_complete:
            continue
        next_batch.append(file)
        num_tasks += len(file.tasks)
        # ordering is important here in case this value is somehow set to 0.
        if num_tasks >= env.MAX_FILES_PER_FINALIZE_BATCH:
            break

    return next_batch


def _start_finalize(c: 'GridRestClient', grid_dir: Path, work: Work, progress_bar: PBar) -> Work:
    """Entrypoint to finalize uploaded files.
    """
    progress_bar.p_finalize.update(progress_bar.finalize_task_id, visible=True)

    while True:
        if done_event.is_set():
            raise RuntimeError(errors.datastore_upload_process_cancelled())

        progress_bar.action_status.update(progress_bar.status_text_finalizing)
        completed_batch = _get_next_finalize_batch(work)
        if len(completed_batch) == 0:
            break

        complete_presigned_url_upload(
            c=c, cluster_id=work.cluster_id, datastore_id=work.datastore_id, data_files=completed_batch
        )
        completed_parts = 0
        for completed_file in completed_batch:
            work.files[completed_file.local_id].is_marked_complete = True
            completed_parts += int(work.files[completed_file.local_id].part_count)

        progress_bar.p_finalize.update(progress_bar.finalize_task_id, advance=completed_parts)
        progress_bar.action_status.update(progress_bar.status_text_saving)
        _save_work_state(grid_dir=grid_dir, work=work)

    return work


# ------------------------- module user methods ---------------------------


@wrap_signals
def resume_datastore_upload(client: 'GridRestClient', grid_dir: Path, work: Work):
    pbar = PBar()
    with Live(pbar.progress_group):
        pbar.action_status.update(pbar.status_text_initializing)
        pbar.setup_progress_meter(work)

        _start_upload(c=client, grid_dir=grid_dir, work=work, progress_bar=pbar)
        _start_finalize(c=client, grid_dir=grid_dir, work=work, progress_bar=pbar)
        mark_datastore_upload_complete(c=client, cluster_id=work.cluster_id, datastore_id=work.datastore_id)
        pbar.action_status.update(pbar.status_text_done)
        remove_datastore_work_state(grid_dir=grid_dir, datastore_id=work.datastore_id)
    return True


@wrap_signals
def begin_new_datastore_upload(
    client: 'GridRestClient',
    grid_dir: Path,
    source_path: Path,
    cluster_id: str,
    datastore_id: str,
    datastore_name: str,
    datastore_version: str,
    creation_timestamp: datetime.datetime,
):
    creation_timestamp = int(time.mktime(creation_timestamp.timetuple()))  # convert to unix time

    pbar = PBar()
    with Live(pbar.progress_group):
        pbar.action_status.update(pbar.status_text_initializing)
        work = initialize_upload_work(
            name=datastore_name,
            version=datastore_version,
            datastore_id=datastore_id,
            cluster_id=cluster_id,
            creation_timestamp=creation_timestamp,
            source_path=str(source_path),
        )
        pbar.setup_progress_meter(work)

        _save_work_state(grid_dir, work)
        _start_upload(c=client, grid_dir=grid_dir, work=work, progress_bar=pbar)
        _start_finalize(c=client, grid_dir=grid_dir, work=work, progress_bar=pbar)
        mark_datastore_upload_complete(c=client, cluster_id=cluster_id, datastore_id=datastore_id)
        pbar.action_status.update(pbar.status_text_done)
        remove_datastore_work_state(grid_dir=grid_dir, datastore_id=datastore_id)

    return True



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


import pprint
import re  # noqa: F401
from typing import TYPE_CHECKING

import six

from grid.openapi.configuration import Configuration

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *

class V1SessionSpec(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'cluster_id': 'str',
        'datastores': 'list[V1DatastoreInput]',
        'desired_state': 'V1SessionState',
        'instance_type': 'str',
        'pause_automation': 'bool',
        'resources': 'V1Resources',
        'use_spot': 'bool',
        'user_id': 'str'
    }

    attribute_map = {
        'cluster_id': 'clusterId',
        'datastores': 'datastores',
        'desired_state': 'desiredState',
        'instance_type': 'instanceType',
        'pause_automation': 'pauseAutomation',
        'resources': 'resources',
        'use_spot': 'useSpot',
        'user_id': 'userId'
    }

    def __init__(self, cluster_id: 'str' = None, datastores: 'list[V1DatastoreInput]' = None, desired_state: 'V1SessionState' = None, instance_type: 'str' = None, pause_automation: 'bool' = None, resources: 'V1Resources' = None, use_spot: 'bool' = None, user_id: 'str' = None, _configuration=None):  # noqa: E501
        """V1SessionSpec - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._cluster_id = None
        self._datastores = None
        self._desired_state = None
        self._instance_type = None
        self._pause_automation = None
        self._resources = None
        self._use_spot = None
        self._user_id = None
        self.discriminator = None

        if cluster_id is not None:
            self.cluster_id = cluster_id
        if datastores is not None:
            self.datastores = datastores
        if desired_state is not None:
            self.desired_state = desired_state
        if instance_type is not None:
            self.instance_type = instance_type
        if pause_automation is not None:
            self.pause_automation = pause_automation
        if resources is not None:
            self.resources = resources
        if use_spot is not None:
            self.use_spot = use_spot
        if user_id is not None:
            self.user_id = user_id

    @property
    def cluster_id(self) -> 'str':
        """Gets the cluster_id of this V1SessionSpec.  # noqa: E501


        :return: The cluster_id of this V1SessionSpec.  # noqa: E501
        :rtype: str
        """
        return self._cluster_id

    @cluster_id.setter
    def cluster_id(self, cluster_id: 'str'):
        """Sets the cluster_id of this V1SessionSpec.


        :param cluster_id: The cluster_id of this V1SessionSpec.  # noqa: E501
        :type: str
        """

        self._cluster_id = cluster_id

    @property
    def datastores(self) -> 'list[V1DatastoreInput]':
        """Gets the datastores of this V1SessionSpec.  # noqa: E501


        :return: The datastores of this V1SessionSpec.  # noqa: E501
        :rtype: list[V1DatastoreInput]
        """
        return self._datastores

    @datastores.setter
    def datastores(self, datastores: 'list[V1DatastoreInput]'):
        """Sets the datastores of this V1SessionSpec.


        :param datastores: The datastores of this V1SessionSpec.  # noqa: E501
        :type: list[V1DatastoreInput]
        """

        self._datastores = datastores

    @property
    def desired_state(self) -> 'V1SessionState':
        """Gets the desired_state of this V1SessionSpec.  # noqa: E501


        :return: The desired_state of this V1SessionSpec.  # noqa: E501
        :rtype: V1SessionState
        """
        return self._desired_state

    @desired_state.setter
    def desired_state(self, desired_state: 'V1SessionState'):
        """Sets the desired_state of this V1SessionSpec.


        :param desired_state: The desired_state of this V1SessionSpec.  # noqa: E501
        :type: V1SessionState
        """

        self._desired_state = desired_state

    @property
    def instance_type(self) -> 'str':
        """Gets the instance_type of this V1SessionSpec.  # noqa: E501


        :return: The instance_type of this V1SessionSpec.  # noqa: E501
        :rtype: str
        """
        return self._instance_type

    @instance_type.setter
    def instance_type(self, instance_type: 'str'):
        """Sets the instance_type of this V1SessionSpec.


        :param instance_type: The instance_type of this V1SessionSpec.  # noqa: E501
        :type: str
        """

        self._instance_type = instance_type

    @property
    def pause_automation(self) -> 'bool':
        """Gets the pause_automation of this V1SessionSpec.  # noqa: E501


        :return: The pause_automation of this V1SessionSpec.  # noqa: E501
        :rtype: bool
        """
        return self._pause_automation

    @pause_automation.setter
    def pause_automation(self, pause_automation: 'bool'):
        """Sets the pause_automation of this V1SessionSpec.


        :param pause_automation: The pause_automation of this V1SessionSpec.  # noqa: E501
        :type: bool
        """

        self._pause_automation = pause_automation

    @property
    def resources(self) -> 'V1Resources':
        """Gets the resources of this V1SessionSpec.  # noqa: E501


        :return: The resources of this V1SessionSpec.  # noqa: E501
        :rtype: V1Resources
        """
        return self._resources

    @resources.setter
    def resources(self, resources: 'V1Resources'):
        """Sets the resources of this V1SessionSpec.


        :param resources: The resources of this V1SessionSpec.  # noqa: E501
        :type: V1Resources
        """

        self._resources = resources

    @property
    def use_spot(self) -> 'bool':
        """Gets the use_spot of this V1SessionSpec.  # noqa: E501


        :return: The use_spot of this V1SessionSpec.  # noqa: E501
        :rtype: bool
        """
        return self._use_spot

    @use_spot.setter
    def use_spot(self, use_spot: 'bool'):
        """Sets the use_spot of this V1SessionSpec.


        :param use_spot: The use_spot of this V1SessionSpec.  # noqa: E501
        :type: bool
        """

        self._use_spot = use_spot

    @property
    def user_id(self) -> 'str':
        """Gets the user_id of this V1SessionSpec.  # noqa: E501


        :return: The user_id of this V1SessionSpec.  # noqa: E501
        :rtype: str
        """
        return self._user_id

    @user_id.setter
    def user_id(self, user_id: 'str'):
        """Sets the user_id of this V1SessionSpec.


        :param user_id: The user_id of this V1SessionSpec.  # noqa: E501
        :type: str
        """

        self._user_id = user_id

    def to_dict(self) -> dict:
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(V1SessionSpec, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self) -> str:
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self) -> str:
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other: 'V1SessionSpec') -> bool:
        """Returns true if both objects are equal"""
        if not isinstance(other, V1SessionSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other: 'V1SessionSpec') -> bool:
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1SessionSpec):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


import pprint
import re  # noqa: F401
from typing import TYPE_CHECKING

import six

from grid.openapi.configuration import Configuration

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *

class V1KubernetesClusterDriver(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'aws': 'V1AWSClusterDriverSpec',
        'azure': 'V1AzureClusterDriverSpec',
        'extra_gridlet_env_var': 'dict(str, str)',
        'gridlet_resources': 'V1ContainerResources',
        'kubeconfig': 'V1ExternalKubeconfig',
        'kubernetes_only_mode': 'bool',
        'namespace': 'str',
        'private_ssh_gateway': 'bool',
        'root_domain_name': 'str'
    }

    attribute_map = {
        'aws': 'aws',
        'azure': 'azure',
        'extra_gridlet_env_var': 'extraGridletEnvVar',
        'gridlet_resources': 'gridletResources',
        'kubeconfig': 'kubeconfig',
        'kubernetes_only_mode': 'kubernetesOnlyMode',
        'namespace': 'namespace',
        'private_ssh_gateway': 'privateSshGateway',
        'root_domain_name': 'rootDomainName'
    }

    def __init__(self, aws: 'V1AWSClusterDriverSpec' = None, azure: 'V1AzureClusterDriverSpec' = None, extra_gridlet_env_var: 'dict(str, str)' = None, gridlet_resources: 'V1ContainerResources' = None, kubeconfig: 'V1ExternalKubeconfig' = None, kubernetes_only_mode: 'bool' = None, namespace: 'str' = None, private_ssh_gateway: 'bool' = None, root_domain_name: 'str' = None, _configuration=None):  # noqa: E501
        """V1KubernetesClusterDriver - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._aws = None
        self._azure = None
        self._extra_gridlet_env_var = None
        self._gridlet_resources = None
        self._kubeconfig = None
        self._kubernetes_only_mode = None
        self._namespace = None
        self._private_ssh_gateway = None
        self._root_domain_name = None
        self.discriminator = None

        if aws is not None:
            self.aws = aws
        if azure is not None:
            self.azure = azure
        if extra_gridlet_env_var is not None:
            self.extra_gridlet_env_var = extra_gridlet_env_var
        if gridlet_resources is not None:
            self.gridlet_resources = gridlet_resources
        if kubeconfig is not None:
            self.kubeconfig = kubeconfig
        if kubernetes_only_mode is not None:
            self.kubernetes_only_mode = kubernetes_only_mode
        if namespace is not None:
            self.namespace = namespace
        if private_ssh_gateway is not None:
            self.private_ssh_gateway = private_ssh_gateway
        if root_domain_name is not None:
            self.root_domain_name = root_domain_name

    @property
    def aws(self) -> 'V1AWSClusterDriverSpec':
        """Gets the aws of this V1KubernetesClusterDriver.  # noqa: E501


        :return: The aws of this V1KubernetesClusterDriver.  # noqa: E501
        :rtype: V1AWSClusterDriverSpec
        """
        return self._aws

    @aws.setter
    def aws(self, aws: 'V1AWSClusterDriverSpec'):
        """Sets the aws of this V1KubernetesClusterDriver.


        :param aws: The aws of this V1KubernetesClusterDriver.  # noqa: E501
        :type: V1AWSClusterDriverSpec
        """

        self._aws = aws

    @property
    def azure(self) -> 'V1AzureClusterDriverSpec':
        """Gets the azure of this V1KubernetesClusterDriver.  # noqa: E501


        :return: The azure of this V1KubernetesClusterDriver.  # noqa: E501
        :rtype: V1AzureClusterDriverSpec
        """
        return self._azure

    @azure.setter
    def azure(self, azure: 'V1AzureClusterDriverSpec'):
        """Sets the azure of this V1KubernetesClusterDriver.


        :param azure: The azure of this V1KubernetesClusterDriver.  # noqa: E501
        :type: V1AzureClusterDriverSpec
        """

        self._azure = azure

    @property
    def extra_gridlet_env_var(self) -> 'dict(str, str)':
        """Gets the extra_gridlet_env_var of this V1KubernetesClusterDriver.  # noqa: E501


        :return: The extra_gridlet_env_var of this V1KubernetesClusterDriver.  # noqa: E501
        :rtype: dict(str, str)
        """
        return self._extra_gridlet_env_var

    @extra_gridlet_env_var.setter
    def extra_gridlet_env_var(self, extra_gridlet_env_var: 'dict(str, str)'):
        """Sets the extra_gridlet_env_var of this V1KubernetesClusterDriver.


        :param extra_gridlet_env_var: The extra_gridlet_env_var of this V1KubernetesClusterDriver.  # noqa: E501
        :type: dict(str, str)
        """

        self._extra_gridlet_env_var = extra_gridlet_env_var

    @property
    def gridlet_resources(self) -> 'V1ContainerResources':
        """Gets the gridlet_resources of this V1KubernetesClusterDriver.  # noqa: E501


        :return: The gridlet_resources of this V1KubernetesClusterDriver.  # noqa: E501
        :rtype: V1ContainerResources
        """
        return self._gridlet_resources

    @gridlet_resources.setter
    def gridlet_resources(self, gridlet_resources: 'V1ContainerResources'):
        """Sets the gridlet_resources of this V1KubernetesClusterDriver.


        :param gridlet_resources: The gridlet_resources of this V1KubernetesClusterDriver.  # noqa: E501
        :type: V1ContainerResources
        """

        self._gridlet_resources = gridlet_resources

    @property
    def kubeconfig(self) -> 'V1ExternalKubeconfig':
        """Gets the kubeconfig of this V1KubernetesClusterDriver.  # noqa: E501


        :return: The kubeconfig of this V1KubernetesClusterDriver.  # noqa: E501
        :rtype: V1ExternalKubeconfig
        """
        return self._kubeconfig

    @kubeconfig.setter
    def kubeconfig(self, kubeconfig: 'V1ExternalKubeconfig'):
        """Sets the kubeconfig of this V1KubernetesClusterDriver.


        :param kubeconfig: The kubeconfig of this V1KubernetesClusterDriver.  # noqa: E501
        :type: V1ExternalKubeconfig
        """

        self._kubeconfig = kubeconfig

    @property
    def kubernetes_only_mode(self) -> 'bool':
        """Gets the kubernetes_only_mode of this V1KubernetesClusterDriver.  # noqa: E501


        :return: The kubernetes_only_mode of this V1KubernetesClusterDriver.  # noqa: E501
        :rtype: bool
        """
        return self._kubernetes_only_mode

    @kubernetes_only_mode.setter
    def kubernetes_only_mode(self, kubernetes_only_mode: 'bool'):
        """Sets the kubernetes_only_mode of this V1KubernetesClusterDriver.


        :param kubernetes_only_mode: The kubernetes_only_mode of this V1KubernetesClusterDriver.  # noqa: E501
        :type: bool
        """

        self._kubernetes_only_mode = kubernetes_only_mode

    @property
    def namespace(self) -> 'str':
        """Gets the namespace of this V1KubernetesClusterDriver.  # noqa: E501


        :return: The namespace of this V1KubernetesClusterDriver.  # noqa: E501
        :rtype: str
        """
        return self._namespace

    @namespace.setter
    def namespace(self, namespace: 'str'):
        """Sets the namespace of this V1KubernetesClusterDriver.


        :param namespace: The namespace of this V1KubernetesClusterDriver.  # noqa: E501
        :type: str
        """

        self._namespace = namespace

    @property
    def private_ssh_gateway(self) -> 'bool':
        """Gets the private_ssh_gateway of this V1KubernetesClusterDriver.  # noqa: E501

        Make gridlet load balancer be within the private subnets. This is cloud specific, thus for AWS the load balancer will be created in the private VPC subnet. The byoc administrator is responsible for VPN solution, that is networking between clients and the gridlet load balancer, that is private subnets on the cloud provider.  # noqa: E501

        :return: The private_ssh_gateway of this V1KubernetesClusterDriver.  # noqa: E501
        :rtype: bool
        """
        return self._private_ssh_gateway

    @private_ssh_gateway.setter
    def private_ssh_gateway(self, private_ssh_gateway: 'bool'):
        """Sets the private_ssh_gateway of this V1KubernetesClusterDriver.

        Make gridlet load balancer be within the private subnets. This is cloud specific, thus for AWS the load balancer will be created in the private VPC subnet. The byoc administrator is responsible for VPN solution, that is networking between clients and the gridlet load balancer, that is private subnets on the cloud provider.  # noqa: E501

        :param private_ssh_gateway: The private_ssh_gateway of this V1KubernetesClusterDriver.  # noqa: E501
        :type: bool
        """

        self._private_ssh_gateway = private_ssh_gateway

    @property
    def root_domain_name(self) -> 'str':
        """Gets the root_domain_name of this V1KubernetesClusterDriver.  # noqa: E501


        :return: The root_domain_name of this V1KubernetesClusterDriver.  # noqa: E501
        :rtype: str
        """
        return self._root_domain_name

    @root_domain_name.setter
    def root_domain_name(self, root_domain_name: 'str'):
        """Sets the root_domain_name of this V1KubernetesClusterDriver.


        :param root_domain_name: The root_domain_name of this V1KubernetesClusterDriver.  # noqa: E501
        :type: str
        """

        self._root_domain_name = root_domain_name

    def to_dict(self) -> dict:
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(V1KubernetesClusterDriver, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self) -> str:
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self) -> str:
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other: 'V1KubernetesClusterDriver') -> bool:
        """Returns true if both objects are equal"""
        if not isinstance(other, V1KubernetesClusterDriver):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other: 'V1KubernetesClusterDriver') -> bool:
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1KubernetesClusterDriver):
            return True

        return self.to_dict() != other.to_dict()



from copy import copy
import hashlib
from pathlib import Path
from typing import Set, Tuple


def dirhash(path: Path,
            algorithm: str = "blake2",
            ignore: Set[str] = {},
            chunk_num_blocks: int = 128) -> Tuple[Set[Path], str]:
    """
    Hashes the contents of an entire directory, excluding known glob patterns.

    Parameters
    ----------
    path: Path
        Path to directory.
    algorithm: str, default "blake2"
        Algorithm to hash contents. "blake2" is set by default because it
        is faster than "md5". [1]
    ignore: Set[str], default {}
        Set of glob patterns to ignore.
    chunk_num_blocks: int, default 128
        Block size to user when iterating over file chunks.

    References
    ----------
    [1] https://crypto.stackexchange.com/questions/70101/blake2-vs-md5-for-checksum-file-integrity
    [2] https://stackoverflow.com/questions/1131220/get-md5-hash-of-big-files-in-python
    """
    # validate input
    if algorithm == "blake2":
        h = hashlib.blake2b(digest_size=20)
    elif algorithm == "md5":
        h = hashlib.md5()
    else:
        raise ValueError(f"Algorithm {algorithm} not supported")

    # make set with absolute file paths to be excluded
    def ignore_glob(path: Path, ignore_pattern: str) -> Set[Path]:
        # If pattern starts with / or ./ glob relative to directory root
        if ignore_pattern.startswith("/"):
            return set(path.glob(ignore_pattern[1:]))
        if ignore_pattern.startswith("./"):
            return set(path.glob(ignore_pattern))
        # All other patterns are recursive. E.g. __pycache__ or *.pyc searches in all sub-directories
        return set(path.rglob(ignore_pattern))

    exclude = {filename for ignore_pattern in ignore for filename in ignore_glob(path, ignore_pattern)}

    # include only files that are not excluded and parent directories are also not excluded
    include = {
        filename
        for filename in path.glob("**/*")
        if filename not in exclude and not any(exclude_path in filename.parents for exclude_path in exclude)
    }

    # calculate hash for all files
    hashed_files = set()

    # sorting preserves order which is required for deterministic hash calculation
    for include_path in sorted(include):
        if not include_path.is_file():
            continue

        hashed_files.add(include_path)
        with include_path.open("rb") as f:
            for chunk in iter(lambda: f.read(chunk_num_blocks * h.block_size), b''):
                h.update(chunk)

    return (hashed_files, h.hexdigest())



from grid.sdk import (
    Datastore,
    Experiment,
    get_instance_types,
    list_datastores,
    list_sessions,
    login,
    Run,
    list_runs,
    Actions,
    ScratchSpace,
    Resources,
    Session,
    Artifact,
    download_artifacts,
    list_artifacts,
)

__all__ = [
    "login", "list_datastores", "Run", "Experiment", "list_runs", "Actions", "ScratchSpace", "Resources", "Datastore",
    "get_instance_types", "list_sessions", "Session", "Artifact", "download_artifacts", "list_artifacts"
]



from functools import lru_cache, wraps
import json
from typing import Dict, Iterator, Optional

from gql import Client, gql
from gql.transport.exceptions import TransportProtocolError
from gql.transport.requests import RequestsHTTPTransport
from gql.transport.websockets import WebsocketsTransport
import requests
from websockets.exceptions import ConnectionClosedError, ConnectionClosedOK

from grid.metadata import __version__
from grid.sdk.auth import Credentials


def humanize_gql_exceptions(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)

        # runtime errors are generated if gql returns with an "error" message
        except RuntimeError as e:
            raise e from None

        except requests.exceptions.ConnectionError:
            raise ConnectionError(
                f'Grid is unreachable. Can you access https://platform.grid.ai ? '
                f'If not, please try again later.'
            ) from None

        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 401:
                raise ConnectionRefusedError('Not authorized. Did you login?') from None
            if e.response.status_code == 500:
                raise ConnectionError('Grid is having issues. Please again later.') from None
            raise ConnectionError(f'We encountered an unknown error. Please try again later. {type(e).__name__}: {e}')

        except requests.exceptions.Timeout:
            raise ConnectionError('The grid platform could not be reached. Are you online?') from None
        except TimeoutError:
            raise RuntimeError(f"Timeout waiting for messages exceeded") from None

        except TransportProtocolError:
            raise ConnectionRefusedError('You are not authorized to use Grid. Did you login?') from None

        # If connection is suddenly closed, indicate that a known error happened.
        except ConnectionClosedError as e:
            raise ConnectionError(f"Interacting with closed connection. Error: {type(e).__name__}") from None
        except ConnectionClosedOK as e:
            raise ConnectionError(
                f"Interacting with properly closed connection. No new "
                f"messages available. Error {type(e).__name__}"
            ) from None

        except Exception as e:
            raise RuntimeError(f"Unhandled error. Please report to the grid team. {type(e).__name__}: {str(e)}")

    return wrapper


@lru_cache(maxsize=None)
@humanize_gql_exceptions
def gql_client(
    url: str,
    creds: Credentials,
    websocket: bool = False,
    *,
    timeout: int = 10,
    headers: Optional[Dict[str, str]] = None
) -> Client:
    """Only create a client instance if there isn't one.

    TODO(rlizzo): re-enable check_version_compatability when versioning is fixed
    """
    if headers is None:
        headers = {}

    headers.update({
        'Content-type': 'application/json',
        'User-Agent': f'grid-api-{__version__}',
        "X-Grid-User": creds.user_id,
        "X-Grid-Key": creds.api_key,
    })

    if websocket:
        _url = url.replace('http://', 'ws://')
        _url = _url.replace('https://', 'wss://')
        _url = _url.replace('graphql', 'subscriptions')
        transport = WebsocketsTransport(url=_url, init_payload=headers)
    else:
        transport = RequestsHTTPTransport(url=url, use_json=True, headers=headers, timeout=timeout, retries=3)

    res = Client(transport=transport, fetch_schema_from_transport=True)
    return res


@humanize_gql_exceptions
def gql_execute(client: Client, query: str, **kwargs) -> dict:
    """Executes a GraphQL query against Grid's API.
    """
    res = client.execute(gql(query), variable_values=kwargs)
    if 'error' in res:
        raise RuntimeError(json.dumps(res['error']))
    return res



from typing import List


def datastore_ask_if_should_resume_upload(source: str) -> str:
    return (
        f"Your previous upload of {source} did not complete. Would you like to complete it? \n"
        f"[y] yes (progress will resume where you left off)\n"
        f"[n] no (progress will be deleted)\n"
    )


def datastore_ask_if_should_resume_after_files_modified(file_list: List[str]) -> str:
    start_text = "The following files have been modified since uploading began: \n"

    file_list_string = ""
    for file in file_list:
        file_list_string += f"- {file}\n"

    end_text = (
        "Would you like to continue uploading the remaining files? Files already "
        "uploaded will not reflect these changes. \n"
        "[y] continue \n"
        "[n] restart (current upload progress will be lost)\n"
    )

    return f"{start_text}{file_list_string}{end_text}"



from dataclasses import asdict, dataclass
from datetime import datetime
from typing import Dict


@dataclass
class ClusterLogsRequest:
    cluster_id: str
    follow: bool
    start: int  # unix timestamp
    end: int  # unix timestamp
    limit: int

    def as_dict(self) -> Dict:
        params = asdict(self)
        # type of the value must be string for downstream url
        # preparation (PreparedRequest().prepare_url()) to work
        params['follow'] = "true" if params.get("follow") else "false"
        return params


@dataclass
class ClusterLogsResponse:
    message: str
    timestamp: datetime
    level: str
    labels: Dict[str, str]

    def as_dict(self) -> Dict:
        return asdict(self)



"""Hook into the low level graphql Client transports and record HTTP request/responses

This allows us to generate a golden file unique to each test invocation against one
of the live dev environments. The query and response is recorded once, after which
future tests are able to just create a mock Client which reads the corresponding
golden files from disk and populates a mapping of request->response as it was
performed in a known good scenario.
"""

import logging
from pathlib import Path

GOLDEN_DIR = Path(__file__).parent.parent.parent / "tests" / "data" / "golden"


# noinspection PyMissingConstructor
class _FilterGQLSchemaQueries(logging.Filter):
    def __init__(self, param=None):
        self.param = param

    def filter(self, record):
        allow = self.param not in record.getMessage()
        record.msg = record.getMessage().replace("\\n", "").strip()
        record.args = ()
        return allow


class _FilterGQLMissingVariablesQuery(logging.Filter):
    def __init__(self, param=None):
        self.param = param

    def filter(self, record):
        msg = record.getMessage()
        if msg.startswith(">>>") and '", "variables": {' not in msg:
            record.msg = record.getMessage()[:-1] + ', "variables": {}}'
            record.args = ()
        return True


class _FilterRestHttpInfoRequest(logging.Filter):
    def __init__(self):  # skipcq: PYL-W0231
        pass

    def filter(self, record):
        msg = record.getMessage()
        if (msg.startswith(">>>") or msg.startswith("<<<")) and '_with_http_info' in msg:
            return False
        return True


_filter_request = _FilterGQLSchemaQueries('>>> {"query": "query IntrospectionQuery')
_filter_request_2 = _FilterGQLMissingVariablesQuery()
_filter_response = _FilterGQLSchemaQueries('<<< {"data":{"__schema":{"queryType"')
_filter_rest_http_request_response = _FilterRestHttpInfoRequest()


class QueryLogging:
    def __init__(self):
        logging.basicConfig(level='ERROR', format="%(name)s:%(message)s ")

        self._lg1 = logging.getLogger("gql.transport.requests")
        self._lg1.addFilter(_filter_response)
        self._lg1.addFilter(_filter_request_2)
        self._lg1.addFilter(_filter_request)

        self._lg2 = logging.getLogger("gql.transport.websockets")
        self._lg2.addFilter(_filter_response)
        self._lg2.addFilter(_filter_request_2)
        self._lg2.addFilter(_filter_request)

        self._lg3 = logging.getLogger("gql.transport.aiohttp")
        self._lg3.addFilter(_filter_response)
        self._lg3.addFilter(_filter_request_2)
        self._lg3.addFilter(_filter_request)

        self._lg4 = logging.getLogger("grid.sdk.rest.client")
        self._lg4.addFilter(_filter_rest_http_request_response)

        self._handler = None

    def __enter__(self):
        if self._handler is None:
            raise ValueError(f'output handler has not been set')
        self._lg1.addHandler(self._handler)
        self._lg2.addHandler(self._handler)
        self._lg3.addHandler(self._handler)
        self._lg4.addHandler(self._handler)
        self._activate()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._lg1.removeHandler(self._handler)
        self._lg2.removeHandler(self._handler)
        self._lg3.removeHandler(self._handler)
        self._lg4.removeHandler(self._handler)
        self._handler = None
        self._deactivate()

    def set_output_file_handler(self, filename: str) -> 'QueryLogging':
        self._handler = logging.FileHandler(filename, mode='w')
        return self

    def _activate(self):
        self._lg1.setLevel(logging.INFO)
        self._lg2.setLevel(logging.INFO)
        self._lg3.setLevel(logging.INFO)
        self._lg4.setLevel(logging.INFO)
        logging.basicConfig(level='INFO', format="%(name)s:%(message)s ")

    def _deactivate(self):
        self._lg1.setLevel(logging.ERROR)
        self._lg2.setLevel(logging.ERROR)
        self._lg3.setLevel(logging.ERROR)
        self._lg4.setLevel(logging.ERROR)
        logging.basicConfig(level='ERROR', format="%(name)s:%(message)s ")



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


import pprint
import re  # noqa: F401
from typing import TYPE_CHECKING

import six

from grid.openapi.configuration import Configuration

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *

class V1GetConfigResponse(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'ga_id': 'str',
        'hotjar_side_id': 'str',
        'sentry_dsn': 'str',
        'stripe_publishable_key': 'str'
    }

    attribute_map = {
        'ga_id': 'gaId',
        'hotjar_side_id': 'hotjarSideId',
        'sentry_dsn': 'sentryDsn',
        'stripe_publishable_key': 'stripePublishableKey'
    }

    def __init__(self, ga_id: 'str' = None, hotjar_side_id: 'str' = None, sentry_dsn: 'str' = None, stripe_publishable_key: 'str' = None, _configuration=None):  # noqa: E501
        """V1GetConfigResponse - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._ga_id = None
        self._hotjar_side_id = None
        self._sentry_dsn = None
        self._stripe_publishable_key = None
        self.discriminator = None

        if ga_id is not None:
            self.ga_id = ga_id
        if hotjar_side_id is not None:
            self.hotjar_side_id = hotjar_side_id
        if sentry_dsn is not None:
            self.sentry_dsn = sentry_dsn
        if stripe_publishable_key is not None:
            self.stripe_publishable_key = stripe_publishable_key

    @property
    def ga_id(self) -> 'str':
        """Gets the ga_id of this V1GetConfigResponse.  # noqa: E501


        :return: The ga_id of this V1GetConfigResponse.  # noqa: E501
        :rtype: str
        """
        return self._ga_id

    @ga_id.setter
    def ga_id(self, ga_id: 'str'):
        """Sets the ga_id of this V1GetConfigResponse.


        :param ga_id: The ga_id of this V1GetConfigResponse.  # noqa: E501
        :type: str
        """

        self._ga_id = ga_id

    @property
    def hotjar_side_id(self) -> 'str':
        """Gets the hotjar_side_id of this V1GetConfigResponse.  # noqa: E501


        :return: The hotjar_side_id of this V1GetConfigResponse.  # noqa: E501
        :rtype: str
        """
        return self._hotjar_side_id

    @hotjar_side_id.setter
    def hotjar_side_id(self, hotjar_side_id: 'str'):
        """Sets the hotjar_side_id of this V1GetConfigResponse.


        :param hotjar_side_id: The hotjar_side_id of this V1GetConfigResponse.  # noqa: E501
        :type: str
        """

        self._hotjar_side_id = hotjar_side_id

    @property
    def sentry_dsn(self) -> 'str':
        """Gets the sentry_dsn of this V1GetConfigResponse.  # noqa: E501


        :return: The sentry_dsn of this V1GetConfigResponse.  # noqa: E501
        :rtype: str
        """
        return self._sentry_dsn

    @sentry_dsn.setter
    def sentry_dsn(self, sentry_dsn: 'str'):
        """Sets the sentry_dsn of this V1GetConfigResponse.


        :param sentry_dsn: The sentry_dsn of this V1GetConfigResponse.  # noqa: E501
        :type: str
        """

        self._sentry_dsn = sentry_dsn

    @property
    def stripe_publishable_key(self) -> 'str':
        """Gets the stripe_publishable_key of this V1GetConfigResponse.  # noqa: E501


        :return: The stripe_publishable_key of this V1GetConfigResponse.  # noqa: E501
        :rtype: str
        """
        return self._stripe_publishable_key

    @stripe_publishable_key.setter
    def stripe_publishable_key(self, stripe_publishable_key: 'str'):
        """Sets the stripe_publishable_key of this V1GetConfigResponse.


        :param stripe_publishable_key: The stripe_publishable_key of this V1GetConfigResponse.  # noqa: E501
        :type: str
        """

        self._stripe_publishable_key = stripe_publishable_key

    def to_dict(self) -> dict:
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(V1GetConfigResponse, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self) -> str:
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self) -> str:
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other: 'V1GetConfigResponse') -> bool:
        """Returns true if both objects are equal"""
        if not isinstance(other, V1GetConfigResponse):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other: 'V1GetConfigResponse') -> bool:
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1GetConfigResponse):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


from __future__ import absolute_import

import re  # noqa: F401
from typing import TYPE_CHECKING, Any

# python 2 and python 3 compatibility library
import six

from grid.openapi.api_client import ApiClient

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *


class ClusterServiceApi(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    Ref: https://github.com/swagger-api/swagger-codegen
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

    def cluster_service_create_cluster(self, body: 'V1CreateClusterRequest', **kwargs) -> 'V1CreateClusterResponse':  # noqa: E501
        """cluster_service_create_cluster  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.cluster_service_create_cluster(body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param V1CreateClusterRequest body: (required)
        :return: V1CreateClusterResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.cluster_service_create_cluster_with_http_info(body, **kwargs)  # noqa: E501
        else:
            (data) = self.cluster_service_create_cluster_with_http_info(body, **kwargs)  # noqa: E501
            return data

    def cluster_service_create_cluster_with_http_info(self, body: 'V1CreateClusterRequest', **kwargs) -> 'V1CreateClusterResponse':  # noqa: E501
        """cluster_service_create_cluster  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.cluster_service_create_cluster_with_http_info(body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param V1CreateClusterRequest body: (required)
        :return: V1CreateClusterResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['body']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method cluster_service_create_cluster" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'body' is set
        if self.api_client.client_side_validation and ('body' not in params or
                                                       params['body'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `body` when calling `cluster_service_create_cluster`")  # noqa: E501

        collection_formats = {}

        path_params = {}

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'body' in params:
            body_params = params['body']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/clusters', 'POST',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1CreateClusterResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def cluster_service_delete_cluster(self, id: 'str', **kwargs) -> 'V1DeleteClusterResponse':  # noqa: E501
        """cluster_service_delete_cluster  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.cluster_service_delete_cluster(id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str id: (required)
        :param bool force:
        :return: V1DeleteClusterResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.cluster_service_delete_cluster_with_http_info(id, **kwargs)  # noqa: E501
        else:
            (data) = self.cluster_service_delete_cluster_with_http_info(id, **kwargs)  # noqa: E501
            return data

    def cluster_service_delete_cluster_with_http_info(self, id: 'str', **kwargs) -> 'V1DeleteClusterResponse':  # noqa: E501
        """cluster_service_delete_cluster  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.cluster_service_delete_cluster_with_http_info(id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str id: (required)
        :param bool force:
        :return: V1DeleteClusterResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['id', 'force']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method cluster_service_delete_cluster" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'id' is set
        if self.api_client.client_side_validation and ('id' not in params or
                                                       params['id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `id` when calling `cluster_service_delete_cluster`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'id' in params:
            path_params['id'] = params['id']  # noqa: E501

        query_params = []
        if 'force' in params:
            query_params.append(('force', params['force']))  # noqa: E501

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/clusters/{id}', 'DELETE',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1DeleteClusterResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def cluster_service_get_cluster(self, id: 'str', **kwargs) -> 'V1GetClusterResponse':  # noqa: E501
        """cluster_service_get_cluster  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.cluster_service_get_cluster(id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str id: (required)
        :return: V1GetClusterResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.cluster_service_get_cluster_with_http_info(id, **kwargs)  # noqa: E501
        else:
            (data) = self.cluster_service_get_cluster_with_http_info(id, **kwargs)  # noqa: E501
            return data

    def cluster_service_get_cluster_with_http_info(self, id: 'str', **kwargs) -> 'V1GetClusterResponse':  # noqa: E501
        """cluster_service_get_cluster  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.cluster_service_get_cluster_with_http_info(id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str id: (required)
        :return: V1GetClusterResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['id']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method cluster_service_get_cluster" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'id' is set
        if self.api_client.client_side_validation and ('id' not in params or
                                                       params['id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `id` when calling `cluster_service_get_cluster`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'id' in params:
            path_params['id'] = params['id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/clusters/{id}', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1GetClusterResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def cluster_service_get_cluster_cloud_provider_account_id(self, id: 'str', **kwargs) -> 'V1GetClusterCloudProviderAccountIDResponse':  # noqa: E501
        """cluster_service_get_cluster_cloud_provider_account_id  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.cluster_service_get_cluster_cloud_provider_account_id(id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str id: (required)
        :return: V1GetClusterCloudProviderAccountIDResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.cluster_service_get_cluster_cloud_provider_account_id_with_http_info(id, **kwargs)  # noqa: E501
        else:
            (data) = self.cluster_service_get_cluster_cloud_provider_account_id_with_http_info(id, **kwargs)  # noqa: E501
            return data

    def cluster_service_get_cluster_cloud_provider_account_id_with_http_info(self, id: 'str', **kwargs) -> 'V1GetClusterCloudProviderAccountIDResponse':  # noqa: E501
        """cluster_service_get_cluster_cloud_provider_account_id  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.cluster_service_get_cluster_cloud_provider_account_id_with_http_info(id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str id: (required)
        :return: V1GetClusterCloudProviderAccountIDResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['id']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method cluster_service_get_cluster_cloud_provider_account_id" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'id' is set
        if self.api_client.client_side_validation and ('id' not in params or
                                                       params['id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `id` when calling `cluster_service_get_cluster_cloud_provider_account_id`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'id' in params:
            path_params['id'] = params['id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/clusters/{id}/cloud-provider-id', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1GetClusterCloudProviderAccountIDResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def cluster_service_list_cluster_instance_types(self, id: 'str', **kwargs) -> 'V1ListClusterInstanceTypesResponse':  # noqa: E501
        """cluster_service_list_cluster_instance_types  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.cluster_service_list_cluster_instance_types(id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str id: (required)
        :return: V1ListClusterInstanceTypesResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.cluster_service_list_cluster_instance_types_with_http_info(id, **kwargs)  # noqa: E501
        else:
            (data) = self.cluster_service_list_cluster_instance_types_with_http_info(id, **kwargs)  # noqa: E501
            return data

    def cluster_service_list_cluster_instance_types_with_http_info(self, id: 'str', **kwargs) -> 'V1ListClusterInstanceTypesResponse':  # noqa: E501
        """cluster_service_list_cluster_instance_types  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.cluster_service_list_cluster_instance_types_with_http_info(id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str id: (required)
        :return: V1ListClusterInstanceTypesResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['id']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method cluster_service_list_cluster_instance_types" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'id' is set
        if self.api_client.client_side_validation and ('id' not in params or
                                                       params['id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `id` when calling `cluster_service_list_cluster_instance_types`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'id' in params:
            path_params['id'] = params['id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/clusters/{id}/instance-types', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1ListClusterInstanceTypesResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def cluster_service_list_clusters(self, **kwargs) -> 'V1ListClustersResponse':  # noqa: E501
        """cluster_service_list_clusters  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.cluster_service_list_clusters(async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param list[str] phase_in:
        :param list[str] phase_not_in:
        :param list[str] cluster_type_in:
        :param list[str] cluster_type_not_in:
        :return: V1ListClustersResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.cluster_service_list_clusters_with_http_info(**kwargs)  # noqa: E501
        else:
            (data) = self.cluster_service_list_clusters_with_http_info(**kwargs)  # noqa: E501
            return data

    def cluster_service_list_clusters_with_http_info(self, **kwargs) -> 'V1ListClustersResponse':  # noqa: E501
        """cluster_service_list_clusters  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.cluster_service_list_clusters_with_http_info(async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param list[str] phase_in:
        :param list[str] phase_not_in:
        :param list[str] cluster_type_in:
        :param list[str] cluster_type_not_in:
        :return: V1ListClustersResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['phase_in', 'phase_not_in', 'cluster_type_in', 'cluster_type_not_in']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method cluster_service_list_clusters" % key
                )
            params[key] = val
        del params['kwargs']

        collection_formats = {}

        path_params = {}

        query_params = []
        if 'phase_in' in params:
            query_params.append(('phaseIn', params['phase_in']))  # noqa: E501
            collection_formats['phaseIn'] = 'multi'  # noqa: E501
        if 'phase_not_in' in params:
            query_params.append(('phaseNotIn', params['phase_not_in']))  # noqa: E501
            collection_formats['phaseNotIn'] = 'multi'  # noqa: E501
        if 'cluster_type_in' in params:
            query_params.append(('clusterTypeIn', params['cluster_type_in']))  # noqa: E501
            collection_formats['clusterTypeIn'] = 'multi'  # noqa: E501
        if 'cluster_type_not_in' in params:
            query_params.append(('clusterTypeNotIn', params['cluster_type_not_in']))  # noqa: E501
            collection_formats['clusterTypeNotIn'] = 'multi'  # noqa: E501

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/clusters', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1ListClustersResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def cluster_service_update_cluster(self, id: 'str', body: 'Body', **kwargs) -> 'V1UpdateClusterResponse':  # noqa: E501
        """cluster_service_update_cluster  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.cluster_service_update_cluster(id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str id: (required)
        :param Body body: (required)
        :return: V1UpdateClusterResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.cluster_service_update_cluster_with_http_info(id, body, **kwargs)  # noqa: E501
        else:
            (data) = self.cluster_service_update_cluster_with_http_info(id, body, **kwargs)  # noqa: E501
            return data

    def cluster_service_update_cluster_with_http_info(self, id: 'str', body: 'Body', **kwargs) -> 'V1UpdateClusterResponse':  # noqa: E501
        """cluster_service_update_cluster  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.cluster_service_update_cluster_with_http_info(id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str id: (required)
        :param Body body: (required)
        :return: V1UpdateClusterResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['id', 'body']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method cluster_service_update_cluster" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'id' is set
        if self.api_client.client_side_validation and ('id' not in params or
                                                       params['id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `id` when calling `cluster_service_update_cluster`")  # noqa: E501
        # verify the required parameter 'body' is set
        if self.api_client.client_side_validation and ('body' not in params or
                                                       params['body'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `body` when calling `cluster_service_update_cluster`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'id' in params:
            path_params['id'] = params['id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'body' in params:
            body_params = params['body']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/clusters/{id}', 'PUT',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1UpdateClusterResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


import pprint
import re  # noqa: F401
from typing import TYPE_CHECKING

import six

from grid.openapi.configuration import Configuration

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *

class Body2(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'objects': 'list[V1DatastoreUploadObject]'
    }

    attribute_map = {
        'objects': 'objects'
    }

    def __init__(self, objects: 'list[V1DatastoreUploadObject]' = None, _configuration=None):  # noqa: E501
        """Body2 - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._objects = None
        self.discriminator = None

        if objects is not None:
            self.objects = objects

    @property
    def objects(self) -> 'list[V1DatastoreUploadObject]':
        """Gets the objects of this Body2.  # noqa: E501

        objects essentially contains the a list of file blobs which we want to generate presigned urls for so they can be uploaded as part of the datastore.  # noqa: E501

        :return: The objects of this Body2.  # noqa: E501
        :rtype: list[V1DatastoreUploadObject]
        """
        return self._objects

    @objects.setter
    def objects(self, objects: 'list[V1DatastoreUploadObject]'):
        """Sets the objects of this Body2.

        objects essentially contains the a list of file blobs which we want to generate presigned urls for so they can be uploaded as part of the datastore.  # noqa: E501

        :param objects: The objects of this Body2.  # noqa: E501
        :type: list[V1DatastoreUploadObject]
        """

        self._objects = objects

    def to_dict(self) -> dict:
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(Body2, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self) -> str:
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self) -> str:
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other: 'Body2') -> bool:
        """Returns true if both objects are equal"""
        if not isinstance(other, Body2):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other: 'Body2') -> bool:
        """Returns true if both objects are not equal"""
        if not isinstance(other, Body2):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


import pprint
import re  # noqa: F401
from typing import TYPE_CHECKING

import six

from grid.openapi.configuration import Configuration

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *

class V1RunActions(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'on_build': 'list[str]',
        'on_build_end': 'list[str]',
        'on_build_start': 'list[str]',
        'on_experiment_end': 'list[str]',
        'on_experiment_start': 'list[str]'
    }

    attribute_map = {
        'on_build': 'onBuild',
        'on_build_end': 'onBuildEnd',
        'on_build_start': 'onBuildStart',
        'on_experiment_end': 'onExperimentEnd',
        'on_experiment_start': 'onExperimentStart'
    }

    def __init__(self, on_build: 'list[str]' = None, on_build_end: 'list[str]' = None, on_build_start: 'list[str]' = None, on_experiment_end: 'list[str]' = None, on_experiment_start: 'list[str]' = None, _configuration=None):  # noqa: E501
        """V1RunActions - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._on_build = None
        self._on_build_end = None
        self._on_build_start = None
        self._on_experiment_end = None
        self._on_experiment_start = None
        self.discriminator = None

        if on_build is not None:
            self.on_build = on_build
        if on_build_end is not None:
            self.on_build_end = on_build_end
        if on_build_start is not None:
            self.on_build_start = on_build_start
        if on_experiment_end is not None:
            self.on_experiment_end = on_experiment_end
        if on_experiment_start is not None:
            self.on_experiment_start = on_experiment_start

    @property
    def on_build(self) -> 'list[str]':
        """Gets the on_build of this V1RunActions.  # noqa: E501


        :return: The on_build of this V1RunActions.  # noqa: E501
        :rtype: list[str]
        """
        return self._on_build

    @on_build.setter
    def on_build(self, on_build: 'list[str]'):
        """Sets the on_build of this V1RunActions.


        :param on_build: The on_build of this V1RunActions.  # noqa: E501
        :type: list[str]
        """

        self._on_build = on_build

    @property
    def on_build_end(self) -> 'list[str]':
        """Gets the on_build_end of this V1RunActions.  # noqa: E501

        commands passed to the image builder which are interpreted as RUN commands in a Dockerfile. Executes after installing dependencies from package manager.  # noqa: E501

        :return: The on_build_end of this V1RunActions.  # noqa: E501
        :rtype: list[str]
        """
        return self._on_build_end

    @on_build_end.setter
    def on_build_end(self, on_build_end: 'list[str]'):
        """Sets the on_build_end of this V1RunActions.

        commands passed to the image builder which are interpreted as RUN commands in a Dockerfile. Executes after installing dependencies from package manager.  # noqa: E501

        :param on_build_end: The on_build_end of this V1RunActions.  # noqa: E501
        :type: list[str]
        """

        self._on_build_end = on_build_end

    @property
    def on_build_start(self) -> 'list[str]':
        """Gets the on_build_start of this V1RunActions.  # noqa: E501

        commands passed to the image builder which are interpreted as RUN commands in a Dockerfile. Executes before installing dependencies from package manager.  # noqa: E501

        :return: The on_build_start of this V1RunActions.  # noqa: E501
        :rtype: list[str]
        """
        return self._on_build_start

    @on_build_start.setter
    def on_build_start(self, on_build_start: 'list[str]'):
        """Sets the on_build_start of this V1RunActions.

        commands passed to the image builder which are interpreted as RUN commands in a Dockerfile. Executes before installing dependencies from package manager.  # noqa: E501

        :param on_build_start: The on_build_start of this V1RunActions.  # noqa: E501
        :type: list[str]
        """

        self._on_build_start = on_build_start

    @property
    def on_experiment_end(self) -> 'list[str]':
        """Gets the on_experiment_end of this V1RunActions.  # noqa: E501

        on_experiment_end allows users to specify commands that need to be executed sequentially after the main experiment process ends. This command will be executed on every experiment that the run creates.  # noqa: E501

        :return: The on_experiment_end of this V1RunActions.  # noqa: E501
        :rtype: list[str]
        """
        return self._on_experiment_end

    @on_experiment_end.setter
    def on_experiment_end(self, on_experiment_end: 'list[str]'):
        """Sets the on_experiment_end of this V1RunActions.

        on_experiment_end allows users to specify commands that need to be executed sequentially after the main experiment process ends. This command will be executed on every experiment that the run creates.  # noqa: E501

        :param on_experiment_end: The on_experiment_end of this V1RunActions.  # noqa: E501
        :type: list[str]
        """

        self._on_experiment_end = on_experiment_end

    @property
    def on_experiment_start(self) -> 'list[str]':
        """Gets the on_experiment_start of this V1RunActions.  # noqa: E501

        on_experiment_start allows users to specify commands that need to be executed sequentially before the main experiment process starts. This command will be executed on every experiment that the run creates.  # noqa: E501

        :return: The on_experiment_start of this V1RunActions.  # noqa: E501
        :rtype: list[str]
        """
        return self._on_experiment_start

    @on_experiment_start.setter
    def on_experiment_start(self, on_experiment_start: 'list[str]'):
        """Sets the on_experiment_start of this V1RunActions.

        on_experiment_start allows users to specify commands that need to be executed sequentially before the main experiment process starts. This command will be executed on every experiment that the run creates.  # noqa: E501

        :param on_experiment_start: The on_experiment_start of this V1RunActions.  # noqa: E501
        :type: list[str]
        """

        self._on_experiment_start = on_experiment_start

    def to_dict(self) -> dict:
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(V1RunActions, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self) -> str:
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self) -> str:
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other: 'V1RunActions') -> bool:
        """Returns true if both objects are equal"""
        if not isinstance(other, V1RunActions):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other: 'V1RunActions') -> bool:
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1RunActions):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


import pprint
import re  # noqa: F401
from typing import TYPE_CHECKING

import six

from grid.openapi.configuration import Configuration

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *

class Externalv1DatastoreSpec(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'cluster_id': 'str',
        'fsx_spec': 'V1DatastoreFSxSpec',
        'size_mib': 'int',
        'source': 'str',
        'source_type': 'V1DatastoreSourceType',
        'uploaded_at': 'datetime',
        'user_id': 'str',
        'version': 'int'
    }

    attribute_map = {
        'cluster_id': 'clusterId',
        'fsx_spec': 'fsxSpec',
        'size_mib': 'sizeMib',
        'source': 'source',
        'source_type': 'sourceType',
        'uploaded_at': 'uploadedAt',
        'user_id': 'userId',
        'version': 'version'
    }

    def __init__(self, cluster_id: 'str' = None, fsx_spec: 'V1DatastoreFSxSpec' = None, size_mib: 'int' = None, source: 'str' = None, source_type: 'V1DatastoreSourceType' = None, uploaded_at: 'datetime' = None, user_id: 'str' = None, version: 'int' = None, _configuration=None):  # noqa: E501
        """Externalv1DatastoreSpec - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._cluster_id = None
        self._fsx_spec = None
        self._size_mib = None
        self._source = None
        self._source_type = None
        self._uploaded_at = None
        self._user_id = None
        self._version = None
        self.discriminator = None

        if cluster_id is not None:
            self.cluster_id = cluster_id
        if fsx_spec is not None:
            self.fsx_spec = fsx_spec
        if size_mib is not None:
            self.size_mib = size_mib
        if source is not None:
            self.source = source
        if source_type is not None:
            self.source_type = source_type
        if uploaded_at is not None:
            self.uploaded_at = uploaded_at
        if user_id is not None:
            self.user_id = user_id
        if version is not None:
            self.version = version

    @property
    def cluster_id(self) -> 'str':
        """Gets the cluster_id of this Externalv1DatastoreSpec.  # noqa: E501


        :return: The cluster_id of this Externalv1DatastoreSpec.  # noqa: E501
        :rtype: str
        """
        return self._cluster_id

    @cluster_id.setter
    def cluster_id(self, cluster_id: 'str'):
        """Sets the cluster_id of this Externalv1DatastoreSpec.


        :param cluster_id: The cluster_id of this Externalv1DatastoreSpec.  # noqa: E501
        :type: str
        """

        self._cluster_id = cluster_id

    @property
    def fsx_spec(self) -> 'V1DatastoreFSxSpec':
        """Gets the fsx_spec of this Externalv1DatastoreSpec.  # noqa: E501


        :return: The fsx_spec of this Externalv1DatastoreSpec.  # noqa: E501
        :rtype: V1DatastoreFSxSpec
        """
        return self._fsx_spec

    @fsx_spec.setter
    def fsx_spec(self, fsx_spec: 'V1DatastoreFSxSpec'):
        """Sets the fsx_spec of this Externalv1DatastoreSpec.


        :param fsx_spec: The fsx_spec of this Externalv1DatastoreSpec.  # noqa: E501
        :type: V1DatastoreFSxSpec
        """

        self._fsx_spec = fsx_spec

    @property
    def size_mib(self) -> 'int':
        """Gets the size_mib of this Externalv1DatastoreSpec.  # noqa: E501


        :return: The size_mib of this Externalv1DatastoreSpec.  # noqa: E501
        :rtype: int
        """
        return self._size_mib

    @size_mib.setter
    def size_mib(self, size_mib: 'int'):
        """Sets the size_mib of this Externalv1DatastoreSpec.


        :param size_mib: The size_mib of this Externalv1DatastoreSpec.  # noqa: E501
        :type: int
        """

        self._size_mib = size_mib

    @property
    def source(self) -> 'str':
        """Gets the source of this Externalv1DatastoreSpec.  # noqa: E501


        :return: The source of this Externalv1DatastoreSpec.  # noqa: E501
        :rtype: str
        """
        return self._source

    @source.setter
    def source(self, source: 'str'):
        """Sets the source of this Externalv1DatastoreSpec.


        :param source: The source of this Externalv1DatastoreSpec.  # noqa: E501
        :type: str
        """

        self._source = source

    @property
    def source_type(self) -> 'V1DatastoreSourceType':
        """Gets the source_type of this Externalv1DatastoreSpec.  # noqa: E501


        :return: The source_type of this Externalv1DatastoreSpec.  # noqa: E501
        :rtype: V1DatastoreSourceType
        """
        return self._source_type

    @source_type.setter
    def source_type(self, source_type: 'V1DatastoreSourceType'):
        """Sets the source_type of this Externalv1DatastoreSpec.


        :param source_type: The source_type of this Externalv1DatastoreSpec.  # noqa: E501
        :type: V1DatastoreSourceType
        """

        self._source_type = source_type

    @property
    def uploaded_at(self) -> 'datetime':
        """Gets the uploaded_at of this Externalv1DatastoreSpec.  # noqa: E501


        :return: The uploaded_at of this Externalv1DatastoreSpec.  # noqa: E501
        :rtype: datetime
        """
        return self._uploaded_at

    @uploaded_at.setter
    def uploaded_at(self, uploaded_at: 'datetime'):
        """Sets the uploaded_at of this Externalv1DatastoreSpec.


        :param uploaded_at: The uploaded_at of this Externalv1DatastoreSpec.  # noqa: E501
        :type: datetime
        """

        self._uploaded_at = uploaded_at

    @property
    def user_id(self) -> 'str':
        """Gets the user_id of this Externalv1DatastoreSpec.  # noqa: E501


        :return: The user_id of this Externalv1DatastoreSpec.  # noqa: E501
        :rtype: str
        """
        return self._user_id

    @user_id.setter
    def user_id(self, user_id: 'str'):
        """Sets the user_id of this Externalv1DatastoreSpec.


        :param user_id: The user_id of this Externalv1DatastoreSpec.  # noqa: E501
        :type: str
        """

        self._user_id = user_id

    @property
    def version(self) -> 'int':
        """Gets the version of this Externalv1DatastoreSpec.  # noqa: E501


        :return: The version of this Externalv1DatastoreSpec.  # noqa: E501
        :rtype: int
        """
        return self._version

    @version.setter
    def version(self, version: 'int'):
        """Sets the version of this Externalv1DatastoreSpec.


        :param version: The version of this Externalv1DatastoreSpec.  # noqa: E501
        :type: int
        """

        self._version = version

    def to_dict(self) -> dict:
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(Externalv1DatastoreSpec, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self) -> str:
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self) -> str:
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other: 'Externalv1DatastoreSpec') -> bool:
        """Returns true if both objects are equal"""
        if not isinstance(other, Externalv1DatastoreSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other: 'Externalv1DatastoreSpec') -> bool:
        """Returns true if both objects are not equal"""
        if not isinstance(other, Externalv1DatastoreSpec):
            return True

        return self.to_dict() != other.to_dict()



from __future__ import annotations

from datetime import timezone, datetime
from typing import Optional

import click
from gql import Client

from grid import list_sessions
from grid.cli.observables.base import BaseObservable
from grid.cli.utilities import get_abs_time_difference, string_format_timedelta


class InteractiveNode(BaseObservable):
    """
    Base observable for a Grid interactive node.

    Parameters
    ----------
    client: Client
        GQL client
    """
    def __init__(self, client: Client):
        self.client = client

        super().__init__(client=client, spinner_load_type="Interactive Sessions")

    def get(self, is_global: Optional[bool] = False):
        self.spinner.start()
        try:
            sessions = list_sessions(include_teams=is_global)
        except Exception as e:
            self.spinner.fail("✘")
            raise click.ClickException(e)

        # Create table with results.
        table_cols = ['Session', 'Status', 'Duration', 'Instance Type', 'URL']
        if is_global:
            table_cols = ['Session', 'Status', 'Duration', 'Created By', 'Instance Type', 'URL']
        table = BaseObservable.create_table(columns=table_cols)

        # If there are no nodes active, add a placeholder row indicating that.
        if len(sessions) == 0:
            table.add_row("None Active.", *[" " for i in range(len(table_cols) - 1)])
            self.spinner.stop()
            self.console.print(table)
            return

        # add data for each row to table
        for session in sessions:
            # Calculate how long the interactive node has been up.
            delta = get_abs_time_difference(datetime.now(timezone.utc), session.created_at)
            duration_str = string_format_timedelta(delta)
            if is_global:
                data = [
                    session.name,  # Session
                    session.status.value,  # Status
                    duration_str,  # Duration
                    session.user.username,  # Created By
                    session.instance_type,  # Instance Type
                    session.jupyter_lab_url,  # URL
                ]
            else:
                data = [
                    session.name,  # Session
                    session.status.value,  # Status
                    duration_str,  # Duration
                    session.instance_type,  # Instance Type
                    session.jupyter_lab_url,  # URL
                ]
            # Replace anything that doesn't have data with a placeholder '-'.
            data = [item if item else "-" for item in data]
            table.add_row(*data)

        self.spinner.ok("✔")
        self.console.print(table)

    def follow(self, is_global: Optional[bool] = False):  # pragma: no cover
        pass



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


import pprint
import re  # noqa: F401
from typing import TYPE_CHECKING

import six

from grid.openapi.configuration import Configuration

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *

class V1DatastoreFile(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'filename': 'str',
        'last_modified': 'datetime',
        'md5_checksum': 'str',
        'size_bytes': 'str',
        'url': 'str'
    }

    attribute_map = {
        'filename': 'filename',
        'last_modified': 'lastModified',
        'md5_checksum': 'md5Checksum',
        'size_bytes': 'sizeBytes',
        'url': 'url'
    }

    def __init__(self, filename: 'str' = None, last_modified: 'datetime' = None, md5_checksum: 'str' = None, size_bytes: 'str' = None, url: 'str' = None, _configuration=None):  # noqa: E501
        """V1DatastoreFile - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._filename = None
        self._last_modified = None
        self._md5_checksum = None
        self._size_bytes = None
        self._url = None
        self.discriminator = None

        if filename is not None:
            self.filename = filename
        if last_modified is not None:
            self.last_modified = last_modified
        if md5_checksum is not None:
            self.md5_checksum = md5_checksum
        if size_bytes is not None:
            self.size_bytes = size_bytes
        if url is not None:
            self.url = url

    @property
    def filename(self) -> 'str':
        """Gets the filename of this V1DatastoreFile.  # noqa: E501


        :return: The filename of this V1DatastoreFile.  # noqa: E501
        :rtype: str
        """
        return self._filename

    @filename.setter
    def filename(self, filename: 'str'):
        """Sets the filename of this V1DatastoreFile.


        :param filename: The filename of this V1DatastoreFile.  # noqa: E501
        :type: str
        """

        self._filename = filename

    @property
    def last_modified(self) -> 'datetime':
        """Gets the last_modified of this V1DatastoreFile.  # noqa: E501


        :return: The last_modified of this V1DatastoreFile.  # noqa: E501
        :rtype: datetime
        """
        return self._last_modified

    @last_modified.setter
    def last_modified(self, last_modified: 'datetime'):
        """Sets the last_modified of this V1DatastoreFile.


        :param last_modified: The last_modified of this V1DatastoreFile.  # noqa: E501
        :type: datetime
        """

        self._last_modified = last_modified

    @property
    def md5_checksum(self) -> 'str':
        """Gets the md5_checksum of this V1DatastoreFile.  # noqa: E501


        :return: The md5_checksum of this V1DatastoreFile.  # noqa: E501
        :rtype: str
        """
        return self._md5_checksum

    @md5_checksum.setter
    def md5_checksum(self, md5_checksum: 'str'):
        """Sets the md5_checksum of this V1DatastoreFile.


        :param md5_checksum: The md5_checksum of this V1DatastoreFile.  # noqa: E501
        :type: str
        """
        if (self._configuration.client_side_validation and
                md5_checksum is not None and not re.search(r'^(?:[A-Za-z0-9+\/]{4})*(?:[A-Za-z0-9+\/]{2}==|[A-Za-z0-9+\/]{3}=)?$', md5_checksum)):  # noqa: E501
            raise ValueError(r"Invalid value for `md5_checksum`, must be a follow pattern or equal to `/^(?:[A-Za-z0-9+\/]{4})*(?:[A-Za-z0-9+\/]{2}==|[A-Za-z0-9+\/]{3}=)?$/`")  # noqa: E501

        self._md5_checksum = md5_checksum

    @property
    def size_bytes(self) -> 'str':
        """Gets the size_bytes of this V1DatastoreFile.  # noqa: E501


        :return: The size_bytes of this V1DatastoreFile.  # noqa: E501
        :rtype: str
        """
        return self._size_bytes

    @size_bytes.setter
    def size_bytes(self, size_bytes: 'str'):
        """Sets the size_bytes of this V1DatastoreFile.


        :param size_bytes: The size_bytes of this V1DatastoreFile.  # noqa: E501
        :type: str
        """

        self._size_bytes = size_bytes

    @property
    def url(self) -> 'str':
        """Gets the url of this V1DatastoreFile.  # noqa: E501


        :return: The url of this V1DatastoreFile.  # noqa: E501
        :rtype: str
        """
        return self._url

    @url.setter
    def url(self, url: 'str'):
        """Sets the url of this V1DatastoreFile.


        :param url: The url of this V1DatastoreFile.  # noqa: E501
        :type: str
        """

        self._url = url

    def to_dict(self) -> dict:
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(V1DatastoreFile, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self) -> str:
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self) -> str:
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other: 'V1DatastoreFile') -> bool:
        """Returns true if both objects are equal"""
        if not isinstance(other, V1DatastoreFile):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other: 'V1DatastoreFile') -> bool:
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1DatastoreFile):
            return True

        return self.to_dict() != other.to_dict()



"""
This file hosts all the websocket service endpoints (which is only cluster logs now). This should be
a python package with cluster/experiment services inside it in the future. Similar to services, we have
models which are the inputs to these services
"""
import json
from typing import Iterator
import logging

import arrow
from websocket import WebSocketConnectionClosedException

from grid.websocket.client import WebSocketClient
from grid.websocket.model import ClusterLogsRequest, ClusterLogsResponse

logger = logging.getLogger(__name__)


class ClusterService:
    service_url = "v1/core/clusters"

    def __init__(self, ws_client: WebSocketClient):
        self.ws_client = ws_client

    def cluster_logs(self, req: ClusterLogsRequest) -> Iterator[ClusterLogsResponse]:
        subscription_path = f"{self.service_url}/{req.cluster_id}/logs"
        query_params = req.as_dict()
        query_params.pop("cluster_id", None)
        ws = self.ws_client.connect(subscription_path, query_params)
        try:
            for message in ws:
                if not message:
                    continue
                kwargs = json.loads(message)
                kwargs['timestamp'] = arrow.get(kwargs['timestamp']).datetime
                kwargs['level'] = kwargs['labels']['level']
                yield ClusterLogsResponse(**kwargs)
        except WebSocketConnectionClosedException:
            logger.debug("WebSocket connection closed")



import functools
import os
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import click
from rich.console import Console
from tqdm import tqdm

from grid.cli import rich_click
from grid.cli.observables.base import create_table
from grid.cli.utilities import is_experiment
from grid.openapi import V1ClusterType, V1GetClusterResponse
from grid.openapi.api import cluster_service_api
from grid.sdk import env
from grid.sdk.artifacts import download_artifacts, list_artifacts
from grid.sdk.client import create_swagger_client
from grid.sdk.experiments import Experiment
from grid.sdk.runs import Run


def with_regular_certs(f):
    """ helper decorator to download artifacts without any custom cert. We can override with USE_CUSTOM_CERTS_ALWAYS (in ../globals.py) """

    # keep track of custom certs
    ssl_custom_cert_envs = "REQUESTS_CA_BUNDLE", "SSL_CERT_FILE"
    custom_certs = dict()

    if not env.SSL_CA_CERT:
        for k in ssl_custom_cert_envs:
            v = os.environ.get(k, None)
            if v is not None:
                custom_certs[k] = v

    @functools.wraps(f)
    def wrapper(*args, **kwargs):

        try:
            for e in custom_certs.keys():
                os.environ.pop(e, None)
            return f(*args, **kwargs)
        finally:
            for e, v in custom_certs.items():
                os.environ[e] = v

    return wrapper


@rich_click.command()
@click.option(
    "--bucket_path",
    "--bucket_paths",
    "bucket_path",
    is_flag=True,
    default=False,
    help="Do not download. Show the bucket URL for the experiments instead. (BYOC only)",
)
@click.option(
    "--download_dir",
    type=click.Path(exists=False, file_okay=False, dir_okay=True),
    required=False,
    default="./grid_artifacts",
    help="Download directory that will host all artifact files."
)
@click.option(
    "-m",
    "--match_regex",
    type=str,
    default="",
    help="Only show artifacts that match this regex filter. Best if quoted."
)
@rich_click.argument(
    "runs_or_experiments", type=str, required=True, nargs=-1, help="The run or experiment to download artifacts for."
)
def artifacts(
    runs_or_experiments: List[str],
    bucket_path: bool = False,
    download_dir: Optional[str] = None,
    match_regex: str = ""
) -> None:
    """Downloads artifacts for a given run or experiments.

    This will download artifacts generated by the runs / experiments.
    Regex filtering is used to determine which artifacts to download.
    """
    experiments: List[Tuple[Experiment, Path]] = []
    for element in runs_or_experiments:
        if is_experiment(element):
            experiment = Experiment(name=element)
            if not experiment.exists:
                click.echo(f"Experiment {element} does not exist - can not get artifacts")
                continue
            experiments.append((experiment, Path(download_dir) / element))

        else:
            run = Run(name=element)
            if not run.exists:
                click.echo(f"Run {element} does not exist - can not get artifacts")
                continue

            for experiment in run.experiments:
                if not experiment.exists:
                    click.echo(f"Experiment {experiment.name} does not exist - can not get artifacts")
                    continue
                experiments.append((experiment, Path(download_dir) / element / experiment.name))

    if len(experiments) == 0:
        click.echo("Aborting download - there are no valid experiments")
        return

    if bucket_path:
        # only valid for BYOC clusters
        api_client = create_swagger_client()
        cluster_api = cluster_service_api.ClusterServiceApi(api_client)
        clusters: Dict[str, V1GetClusterResponse] = {exp.cluster_id: None for exp, _ in experiments}
        for cluster_id in clusters.keys():
            resp = cluster_api.cluster_service_get_cluster(cluster_id)
            if resp.spec.cluster_type != V1ClusterType.BYOC:
                raise click.ClickException(
                    "--bucket_path is only valid for experiments on BYOC clusters. "
                    f"The experiment(s) requested belong to cluster '{cluster_id}' which is not BYOC."
                )
            clusters[cluster_id] = resp

        # print the s3 uri for the experiment artifacts
        table = create_table(['experiment name', 'bucket path to artifacts'])
        for exp, _ in experiments:
            driver = clusters[exp.cluster_id].spec.driver.kubernetes
            if not driver.aws:
                raise click.ClickException("only aws clusters are supported")
            aws = driver.aws
            table.add_row(exp.name, f"s3://{aws.bucket_name}/experiments/{exp.id}/artifacts/")

        console = Console()
        console.print(table)
        return

    click.echo("Downloading artifacts. This command may take a while")
    # exp.download_artifacts generates two progress bars:
    # 1st: progress of all artifacts in an experiment
    # 2nd: download progress for a single artifact in MB/s
    # This will add a 3rd at the bottom: experiments finished downloading
    exp_iter = tqdm(experiments, unit="experiment", position=2)
    for experiment, target_path in exp_iter:
        exp_iter.set_description(experiment.name)
        artifacts_to_download = list_artifacts(experiment.name, artifact_regex=match_regex)
        download_artifacts(artifacts_to_download, destination=target_path)
    click.echo("Done downloading artifacts!")



import abc
from typing import List

from rich.console import Console
from rich.table import Table
from yaspin import yaspin


class BaseObservable(abc.ABC):
    def __init__(self, client=None, spinner_load_type=""):
        self.client = client
        self.console = Console()
        self.spinner = yaspin(text=f"Loading {spinner_load_type}...", color="yellow")

    @abc.abstractmethod
    def get(self, is_global: bool):
        """
        Get the status of the resources

        Parameters
        ----------
        is_global:
            If True, returns status of resources for everyone in the team
        """

    @abc.abstractmethod
    def follow(self, is_global: bool):
        """
        Parameters
        ----------
        is_global:
            If True, returns status of resources for everyone in the team
        """

    @staticmethod
    def create_table(columns: List[str]) -> Table:
        return create_table(columns)


def create_table(columns: List[str]) -> Table:
    table = Table(show_header=True, header_style="bold green")
    table.add_column(columns[0], style='dim')
    for column in columns[1:]:
        table.add_column(column, justify='right')
    return table
